import os, logging
from foresight.environment.git.git_helper import GitHelper
from foresight.environment.git.git_env_info_provider import GitEnvironmentInfoProvider
from foresight.environment.github.github_environment_info_provider import GithubEnvironmentInfoProvider
from foresight.environment.gitlab.gitlab_environment_info_provider import GitlabEnvironmentInfoProvider
from foresight.environment.jenkins.jenkins_environment_info_provider import JenkinsEnvironmentInfoProvider
from foresight.environment.travisci.travisci_environment_info_provider import TravisCIEnvironmentInfoProvider
from foresight.environment.circleci.circleci_environment_info_provider import CircleCIEnvironmentInfoProvider
from foresight.environment.bitbucket.bitbucket_environment_info_provider import BitbucketEnvironmentInfoProvider
from foresight.environment.azure.azure_environment_info_provider import AzureEnvironmentInfoProvider
from foresight.test_runner_tags import TestRunnerTags
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class EnvironmentSupport:

    ENVIRONMENTS_VARS = {
        "GITHUB_ENV": GithubEnvironmentInfoProvider, # https://docs.github.com/en/actions/learn-github-actions/environment-variables
        "CI_PROJECT_ID": GitlabEnvironmentInfoProvider, # https://docs.gitlab.com/ee/ci/variables/predefined_variables.html
        "JENKINS_HOME": JenkinsEnvironmentInfoProvider, # https://e.printstacktrace.blog/jenkins-pipeline-environment-variables-the-definitive-guide/
        "JENKINS_URL": JenkinsEnvironmentInfoProvider, # https://e.printstacktrace.blog/jenkins-pipeline-environment-variables-the-definitive-guide/
        "TRAVIS": TravisCIEnvironmentInfoProvider, # https://docs.travis-ci.com/user/environment-variables/#default-environment-variables
        "CIRCLECI": CircleCIEnvironmentInfoProvider, # https://circleci.com/docs/2.0/env-vars/#built-in-environment-variables
        "BITBUCKET_GIT_HTTP_ORIGIN": BitbucketEnvironmentInfoProvider, # https://support.atlassian.com/bitbucket-cloud/docs/variables-and-secrets/
        "BITBUCKET_GIT_SSH_ORIGIN": BitbucketEnvironmentInfoProvider, # https://support.atlassian.com/bitbucket-cloud/docs/variables-and-secrets/
        "TF_BUILD": AzureEnvironmentInfoProvider # https://docs.microsoft.com/en-us/azure/devops/pipelines/build/variables?view=azure-devops&tabs=yaml
    }
    environment_info = None


    @classmethod
    def init(cls):
        """First check git provider, then iterate over ENVIRONMENTS_VARS dict.
        """
        try:
            LOGGER.debug("Checking ci environments...")
            for key, clz in cls.ENVIRONMENTS_VARS.items():
                LOGGER.debug("Current key, clz: {}, {}".format(key,clz))
                if os.getenv(key):
                    ei = clz.build_env_info()
                    if ei: 
                        cls.environment_info = ei
                        print_debug_message_to_console("Environment info: {}".format(cls.environment_info.to_json()))
                        print_debug_message_to_console("Founded key and class: {}, {}".format(key, clz))
                        break
            if cls.environment_info == None:
                if GitHelper.get_repo_url():
                    print_debug_message_to_console("Couldn't find any ci envrionment! Trying .git file...")
                    cls.environment_info = GitEnvironmentInfoProvider.build_env_info()
                    print_debug_message_to_console("Environment info: {}".format(cls.environment_info.to_json()))
                else:
                    print_debug_message_to_console("Couldn't find .git file!")
        except Exception as err:
            LOGGER.error("Environment Support environment_info could not set: {}".format(err))
            cls.environment_info = None
            print_debug_message_to_console("environment_info is None!")
            pass


    @classmethod
    def set_span_tags(cls, span):
        """Set span data tags corresponds to TestRunner

        Args:
            obj (ThundraSpan): Span or invocation data
        """
        try:
            if cls.environment_info:
                span.set_tag(TestRunnerTags.TEST_ENVIRONMENT, cls.environment_info.environment)
                span.set_tag(TestRunnerTags.SOURCE_CODE_REPO_URL, cls.environment_info.repo_url)
                span.set_tag(TestRunnerTags.SOURCE_CODE_REPO_NAME, cls.environment_info.repo_name)
                span.set_tag(TestRunnerTags.SOURCE_CODE_BRANCH, cls.environment_info.branch)
                span.set_tag(TestRunnerTags.SOURCE_CODE_COMMIT_HASH, cls.environment_info.commit_hash)
                span.set_tag(TestRunnerTags.SOURCE_CODE_COMMIT_MESSAGE, cls.environment_info.commit_message)
        except Exception as e:
            LOGGER.error("Couldn't set environment info for span tags.")
            pass


    @classmethod
    def set_invocation_tags(cls, invocation_data):
        """Set invocation data tags corresponds to TestRunner

        Args:
            obj (invocation): Span or invocation data
        """
        try:
            if cls.environment_info:
                invocation_data[TestRunnerTags.TEST_ENVIRONMENT] =  cls.environment_info.environment
                invocation_data[TestRunnerTags.SOURCE_CODE_REPO_URL] =  cls.environment_info.repo_url
                invocation_data[TestRunnerTags.SOURCE_CODE_REPO_NAME] =  cls.environment_info.repo_name
                invocation_data[TestRunnerTags.SOURCE_CODE_BRANCH] =  cls.environment_info.branch
                invocation_data[TestRunnerTags.SOURCE_CODE_COMMIT_HASH] =  cls.environment_info.commit_hash
                invocation_data[TestRunnerTags.SOURCE_CODE_COMMIT_MESSAGE] =  cls.environment_info.commit_message
        except Exception as e:
            LOGGER.error("Couldn't set environment info for invocation tags.")
            pass


import json


class JSONEncoder(json.JSONEncoder):
    def default(self, z):
        try:
            if isinstance(z, bytes):
                return z.decode('utf-8', errors='ignore')
            elif "to_json" in dir(z):
                return z.to_json()
            else:
                return super(JSONEncoder, self).default(z)
        except Exception as e:
            print(e)


def to_json(data, separators=None):
    return json.dumps(data, separators=separators, cls=JSONEncoder)




def bytes_to_str(value):
    """Convert byte to string

    Args:
        value (byte]):  

    Returns:
        [str]: value if value is str o.w str(value)
    """
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return value


def extract_headers(connection_obj):
    """Convert nested list headers in request/response object to dict

    Args:
        connection_obj (obj): request or response object

    Returns:
        dict: request or response headers dict version
    """
    headers = connection_obj.get("headers")
    if headers:
        return dict((bytes_to_str(k), bytes_to_str(v)) for (k,v) in headers)
    return {}


import copy

import opentracing


class ThundraSpanContext(opentracing.SpanContext):

    def __init__(self,
                 trace_id=None,
                 transaction_id=None,
                 span_id=None,
                 parent_span_id=None,
                 baggage=None):
        self._trace_id = trace_id
        self._transaction_id = transaction_id
        self._span_id = span_id
        self._parent_span_id = parent_span_id
        self._baggage = baggage or opentracing.SpanContext.EMPTY_BAGGAGE

    @property
    def trace_id(self):
        return self._trace_id

    @property
    def transaction_id(self):
        return self._transaction_id

    @property
    def span_id(self):
        return self._span_id

    @span_id.setter
    def span_id(self, value):
        self._span_id = value

    @property
    def parent_span_id(self):
        return self._parent_span_id

    @property
    def baggage(self):
        return self._baggage

    def context_with_baggage_item(self, key, value):
        new_baggage_item = copy.copy(self.baggage)
        new_baggage_item[key] = value
        return ThundraSpanContext(trace_id=self.trace_id,
                                  transaction_id=self.transaction_id,
                                  span_id=self.span_id,
                                  parent_span_id=self.parent_span_id,
                                  baggage=new_baggage_item)



import copy
import logging
import os
import subprocess
import time
import traceback
from functools import wraps

from thundra import constants
from thundra.application.global_application_info_provider import GlobalApplicationInfoProvider
from thundra.compat import PY2, TimeoutError
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.context.global_execution_context_provider import GlobalExecutionContextProvider
from thundra.context.plugin_context import PluginContext
from thundra.integrations import handler_wrappers
from thundra.plugins.log.thundra_logger import debug_logger
from thundra.timeout import Timeout
from thundra.wrappers import wrapper_utils
from thundra.wrappers.aws_lambda import LambdaApplicationInfoProvider
from thundra.wrappers.aws_lambda import lambda_executor
from thundra.wrappers.base_wrapper import BaseWrapper

logger = logging.getLogger(__name__)


class LambdaWrapper(BaseWrapper):

    def __init__(self, api_key=None, disable_trace=False, disable_metric=True, disable_log=True, opts=None):
        super(LambdaWrapper, self).__init__(api_key, disable_trace, disable_metric, disable_log, opts)
        self.application_info_provider = GlobalApplicationInfoProvider(LambdaApplicationInfoProvider())
        self.plugin_context = PluginContext(application_info=self.application_info_provider.get_application_info(),
                                            request_count=0,
                                            executor=lambda_executor,
                                            api_key=self.api_key)

        ExecutionContextManager.set_provider(GlobalExecutionContextProvider())
        self.plugins = wrapper_utils.initialize_plugins(self.plugin_context, disable_trace, disable_metric, disable_log,
                                                        self.config)

        self.timeout_margin = ConfigProvider.get(config_names.THUNDRA_LAMBDA_TIMEOUT_MARGIN,
                                                 constants.DEFAULT_LAMBDA_TIMEOUT_MARGIN)

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INSTRUMENT_DISABLE):
            # Pass thundra instance to integration for wrapping handler wrappers
            handler_wrappers.patch_modules(self)

        self.ptvsd_imported = False
        if ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_ENABLE,
                              ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_AUTH_TOKEN)):
            self.initialize_debugger()

    def __call__(self, original_func):
        if hasattr(original_func, "_thundra_wrapped") or ConfigProvider.get(config_names.THUNDRA_DISABLE, False):
            return original_func

        @wraps(original_func)
        def wrapper(event, context):
            application_name = self.plugin_context.application_info.get('applicationName')
            self.application_info_provider.update({
                'applicationId': LambdaApplicationInfoProvider.get_application_id(context,
                                                                                  application_name=application_name)
            })

            # Execution context initialization
            execution_context = wrapper_utils.create_execution_context()
            try:
                execution_context.platform_data['originalEvent'] = copy.deepcopy(event)
            except:
                execution_context.platform_data['originalEvent'] = event
            execution_context.platform_data['originalContext'] = context
            ExecutionContextManager.set(execution_context)

            # Before running user's handler
            try:
                if ConfigProvider.get(config_names.THUNDRA_LAMBDA_WARMUP_WARMUPAWARE,
                                      False) and self.check_and_handle_warmup_request(event):
                    return None

                self.plugin_context.request_count += 1
                self.execute_hook('before:invocation', execution_context)

                timeout_duration = self.get_timeout_duration(context)
            except Exception as e:
                logger.error("Error during the before part of Thundra: {}".format(e))
                return original_func(event, context)

            # Invoke user handler
            try:
                response = None
                with Timeout(timeout_duration, self.timeout_handler, execution_context):
                    if ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_ENABLE,
                                          ConfigProvider.get(
                                              config_names.THUNDRA_LAMBDA_DEBUGGER_AUTH_TOKEN)) and self.ptvsd_imported:
                        self.start_debugger_tracing(context)

                    response = original_func(event, context)
                    execution_context.response = response
            except Exception as e:
                try:
                    execution_context.error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc()
                    }
                    self.prepare_and_send_reports(execution_context)
                except Exception as e_in:
                    logger.error("Error during the after part of Thundra: {}".format(e_in))
                    pass
                raise e
            finally:
                if ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_ENABLE,
                                      ConfigProvider.get(
                                          config_names.THUNDRA_LAMBDA_DEBUGGER_AUTH_TOKEN)) and self.ptvsd_imported:
                    self.stop_debugger_tracing()

            # After having run the user's handler
            try:
                self.prepare_and_send_reports(execution_context)
            except Exception as e:
                logger.error("Error during the after part of Thundra: {}".format(e))

            ExecutionContextManager.clear()
            return response

        setattr(wrapper, '_thundra_wrapped', True)
        return wrapper

    call = __call__

    def initialize_debugger(self):
        if PY2:
            logger.error("Online debugging not supported in python2.7. Supported versions: 3.6, 3.7, 3.8")
            return
        try:
            import ptvsd
            self.ptvsd_imported = True
        except Exception as e:
            logger.error("Could not import ptvsd. Thundra ptvsd layer must be added")

    def start_debugger_tracing(self, context):
        try:
            import ptvsd
            ptvsd.tracing(True)

            ptvsd.enable_attach(address=("localhost", ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_PORT)))
            if not self.debugger_process:
                env = os.environ.copy()
                env['BROKER_HOST'] = str(ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_BROKER_HOST))
                env['BROKER_PORT'] = str(ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_BROKER_PORT))
                env['DEBUGGER_PORT'] = str(ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_PORT))
                env['AUTH_TOKEN'] = str(ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_AUTH_TOKEN))
                env['SESSION_NAME'] = str(ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_SESSION_NAME))

                if hasattr(context, 'get_remaining_time_in_millis'):
                    env['SESSION_TIMEOUT'] = str(context.get_remaining_time_in_millis() + int(time.time() * 1000.0))

                debug_bridge_file_path = os.path.join(os.path.dirname(__file__), '../../debug/bridge.py')
                self.debugger_process = subprocess.Popen(["python", debug_bridge_file_path], stdout=subprocess.PIPE,
                                                         stdin=subprocess.PIPE, env=env)

            start_time = time.time()
            debug_process_running = True
            while time.time() < (start_time + ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_WAIT_MAX) / 1000) \
                    and not ptvsd.is_attached():
                if self.debugger_process.poll() is None:
                    ptvsd.wait_for_attach(0.01)
                else:
                    debug_process_running = False
                    break

            if not ptvsd.is_attached():
                if debug_process_running:
                    logger.error('Couldn\'t complete debugger handshake in {} milliseconds.' \
                                 .format(ConfigProvider.get(config_names.THUNDRA_LAMBDA_DEBUGGER_WAIT_MAX)))
                ptvsd.tracing(False)
            else:
                ptvsd.tracing(True)

        except Exception as e:
            logger.error("error while setting tracing true to debugger using ptvsd: {}".format(e))

    def stop_debugger_tracing(self):
        try:
            import ptvsd
            ptvsd.tracing(False)
            from ptvsd.attach_server import debugger_attached
            debugger_attached.clear()
        except Exception as e:
            logger.error("error while setting tracing false to debugger using ptvsd: {}".format(e))

        try:
            if self.debugger_process:
                o, e = self.debugger_process.communicate(b"fin\n")
                debug_logger("Thundra debugger process output: {}".format(o.decode("utf-8")))
                self.debugger_process = None
        except Exception as e:
            self.debugger_process = None
            logger.error("error while killing proxy process for debug: {}".format(e))

    def check_and_handle_warmup_request(self, event):

        # Check whether it is empty request which is used as default warmup request
        if not event:
            print("Received warmup request as empty message. " +
                  "Handling with 90 milliseconds delay ...")
            time.sleep(0.1)
            return True
        else:
            if isinstance(event, str):
                # Check whether it is warmup request
                if event.startswith('#warmup'):
                    delayTime = 90
                    args = event[len('#warmup'):].strip().split()
                    # Warmup messages are in '#warmup wait=<waitTime>' format
                    # Iterate over all warmup arguments
                    for arg in args:
                        argParts = arg.split('=')
                        # Check whether argument is in key=value format
                        if len(argParts) == 2:
                            argName = argParts[0]
                            argValue = argParts[1]
                            # Check whether argument is "wait" argument
                            # which specifies extra wait time before returning from request
                            if argName == 'wait':
                                waitTime = int(argValue)
                                delayTime += waitTime
                    print("Received warmup request as warmup message. " +
                          "Handling with " + str(delayTime) + " milliseconds delay ...")
                    time.sleep(delayTime / 1000)
                    return True
            return False

    def get_timeout_duration(self, context):
        timeout_duration = 0
        if hasattr(context, 'get_remaining_time_in_millis'):
            timeout_duration = context.get_remaining_time_in_millis() - self.timeout_margin
            if timeout_duration <= 0:
                timeout_duration = context.get_remaining_time_in_millis() - \
                                   constants.DEFAULT_LAMBDA_TIMEOUT_MARGIN
                logger.warning('Given timeout margin is bigger than lambda timeout duration and '
                               'since the difference is negative, it is set to default value (' +
                               str(constants.DEFAULT_LAMBDA_TIMEOUT_MARGIN) + ')')

        return timeout_duration / 1000.0

    def timeout_handler(self, execution_context):
        execution_context.timeout = True
        execution_context.error = TimeoutError('Task timed out')
        self.prepare_and_send_reports(execution_context)



import abc, sys

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider

ABC = abc.ABCMeta('ABC', (object,), {})


class ApplicationInfoProvider(ABC):
    
    APPLICATION_RUNTIME = "python"
    APPLICATION_RUNTIME_VERSION = str(sys.version_info[0])

    @abc.abstractmethod
    def get_application_info(self):
        pass

    @staticmethod
    def parse_application_tags():
        application_tags = {}
        prefix_length = len(config_names.THUNDRA_APPLICATION_TAG_PREFIX) + 1
        for key in ConfigProvider.configs:
            if key.startswith(config_names.THUNDRA_APPLICATION_TAG_PREFIX) and len(key) >= prefix_length:
                app_tag_key = key[prefix_length:]
                val = ConfigProvider.get(key)
                application_tags[app_tag_key] = val
        return application_tags



import wrapt

from thundra import utils, constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider

def _wrapper(wrapped, instance, args, kwargs):
    """Set middleware to trace Fast api. Fastapi has been built on starlett and pydantic frameworks.
    Request and response flow has been handled by starlette that is a lightweight ASGI framework. Fastapi 
    has an class called APIRouter that extends starlett Router class which is used to handle connections by starlette.
    The middleware should be an ASGI Middleware. Thus, the __call__(scope, receive, send) function should be implemented. 
    By default, starlette add two middleware except user defined middlewares which are ServerErrorMiddleware and ExceptionMiddleware. 
    Middleware list seems like [ServerErrorMiddleware, user_defined_middlewares, ExceptionMiddleware]. This list added in
    reversed order. Therefore, when we add our FastapiMiddleware to the zero index, it is placed top of middleware hierarchy.

    Args:
        wrapped (module): Wrapped module
        instance (function): Module enter point
        args (list): Wrapped function list of arguments
        kwargs (dict): Wrapped function key:value arguments
    """
    from fastapi.middleware import Middleware
    from thundra.wrappers.fastapi.middleware import ThundraMiddleware
    from thundra.wrappers.fastapi.fastapi_wrapper import FastapiWrapper
    middlewares = kwargs.pop("middleware", [])
    middlewares.insert(0, Middleware(ThundraMiddleware, wrapper=FastapiWrapper.get_instance()))
    kwargs.update({"middleware": middlewares})
    wrapped(*args, **kwargs)


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_FASTAPI_DISABLE) and \
        not utils.get_env_variable(constants.AWS_LAMBDA_FUNCTION_NAME):
        wrapt.wrap_function_wrapper(
            "fastapi.applications",
            "FastAPI.__init__",
            _wrapper
        )


import logging
import traceback
from functools import wraps

from thundra import constants, configure
from thundra.application.global_application_info_provider import GlobalApplicationInfoProvider
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.context.plugin_context import PluginContext
from thundra.context.tracing_execution_context_provider import TracingExecutionContextProvider
from thundra.wrappers import wrapper_utils, web_wrapper_utils
from thundra.wrappers.base_wrapper import BaseWrapper
from thundra.wrappers.django import django_executor

logger = logging.getLogger(__name__)


class DjangoWrapper(BaseWrapper):

    def __init__(self, api_key=None, disable_trace=False, disable_metric=True, disable_log=True, opts=None):
        try:
            from django.conf import settings
            django_settings = getattr(settings, 'THUNDRA', None)
            if django_settings is not None:
                configure({'config': django_settings})
        except:
            pass

        super(DjangoWrapper, self).__init__(api_key, disable_trace, disable_metric, disable_log, opts)
        self.application_info_provider = GlobalApplicationInfoProvider()
        ExecutionContextManager.set_provider(TracingExecutionContextProvider())
        self.plugin_context = PluginContext(application_info=self.application_info_provider.get_application_info(),
                                            request_count=0,
                                            executor=django_executor,
                                            api_key=self.api_key)

        self.plugins = wrapper_utils.initialize_plugins(self.plugin_context, disable_trace, disable_metric, disable_log,
                                                        config=self.config)

        web_wrapper_utils.update_application_info(self.application_info_provider, self.plugin_context.application_info,
                                                  constants.ClassNames['DJANGO'])

    def before_request(self, request):
        # Execution context initialization
        execution_context = wrapper_utils.create_execution_context()
        execution_context.platform_data['request'] = request

        # Execute plugin hooks before running user's handler
        self.plugin_context.request_count += 1
        self.execute_hook('before:invocation', execution_context)

        return execution_context

    def after_request(self, response):
        execution_context = ExecutionContextManager.get()
        if response:
            execution_context.response = response
        self.prepare_and_send_reports_async(execution_context)

    def process_exception(self, exception):
        execution_context = ExecutionContextManager.get()
        execution_context.error = exception

    def __call__(self, original_func):
        if hasattr(original_func, "_thundra_wrapped") or ConfigProvider.get(config_names.THUNDRA_DISABLE, False):
            return original_func

        @wraps(original_func)
        def wrapper(request, *args, **kwargs):
            if getattr(request, '_thundra_wrapped', False):
                return original_func(request, *args, **kwargs)
            setattr(request, '_thundra_wrapped', True)
            try:
                execution_context = self.before_request(request)
            except Exception as e:
                logger.error("Error during the before part of Thundra: {}".format(e))
                return original_func(request, *args, **kwargs)

            response = None
            # Invoke user handler
            try:
                response = original_func(request, *args, **kwargs)
            except Exception as e:
                try:
                    execution_context.error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc()
                    }
                    self.after_request(response)
                except Exception as e_in:
                    logger.error("Error during the after part of Thundra: {}".format(e_in))
                raise e

            try:
                self.after_request(response)
            except Exception as e:
                logger.error("Error during the after part of Thundra: {}".format(e))
            return response

        setattr(wrapper, '_thundra_wrapped', True)
        return wrapper

    call = __call__



import foresight.utils.generic_utils as utils

class ForesightConstants:
    TEST_APP_INSTANCE_ID_PREFIX = utils.create_uuid4() + ":"
    TEST_APP_STAGE = "test"
    TEST_FIXTURE_DOMAIN_NAME = "TestFixture"
    TEST_DOMAIN_NAME = "Test"
    TEST_SUITE_DOMAIN_NAME = "TestSuite"
    TEST_OPERATION_NAME = "RunTest"    
    MAX_TEST_METHOD_NAME = 100



import math
import uuid

from thundra import utils, constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.plugins.invocation import invocation_trace_support
from thundra.plugins.log.thundra_logger import debug_logger
from thundra.wrappers import wrapper_utils
from thundra.wrappers.aws_lambda import lambda_event_utils


def start_trace(plugin_context, execution_context, tracer):
    context = execution_context.platform_data['originalContext']
    execution_context.trace_id = str(uuid.uuid4())
    # Start root span
    scope = tracer.start_active_span(operation_name=context.function_name,
                                     start_time=execution_context.start_timestamp,
                                     finish_on_close=False,
                                     trace_id=execution_context.trace_id,
                                     transaction_id=execution_context.transaction_id,
                                     execution_context=execution_context)
    root_span = scope.span

    # Set root span class and domain names
    root_span.class_name = constants.ClassNames['LAMBDA']
    root_span.domain_name = constants.DomainNames['API']
    # Add root span tags
    execution_context.span_id = root_span.context.span_id
    root_span.set_tag('aws.region', utils.get_aws_region_from_arn(
        getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None)))
    root_span.set_tag('aws.lambda.name', context.function_name)
    root_span.set_tag('aws.lambda.arn', getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None))
    root_span.set_tag('aws.lambda.memory.limit', getattr(context, constants.CONTEXT_MEMORY_LIMIT_IN_MB, None))
    root_span.set_tag('aws.lambda.log_group_name', getattr(context, constants.CONTEXT_LOG_GROUP_NAME, None))
    root_span.set_tag('aws.lambda.log_stream_name', getattr(context, constants.CONTEXT_LOG_STREAM_NAME, None))
    root_span.set_tag('aws.lambda.invocation.coldstart', plugin_context.request_count == 1)
    root_span.set_tag('aws.lambda.invocation.timeout', execution_context.timeout)
    root_span.set_tag('aws.lambda.invocation.request_id',
                      getattr(context, constants.CONTEXT_AWS_REQUEST_ID, None))
    inject_trigger_tags(root_span, execution_context.platform_data['originalEvent'], context)

    root_span.on_started()
    execution_context.root_span = root_span
    execution_context.scope = scope


def inject_trigger_tags(span, original_event, original_context):
    try:
        lambda_event_utils.extract_trace_link_from_event(original_event)
        lambda_event_type = lambda_event_utils.get_lambda_event_type(original_event, original_context)

        if lambda_event_type == lambda_event_utils.LambdaEventType.Kinesis:
            lambda_event_utils.inject_trigger_tags_for_kinesis(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.Firehose:
            lambda_event_utils.inject_trigger_tags_for_firehose(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.DynamoDB:
            lambda_event_utils.inject_trigger_tags_for_dynamodb(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.SNS:
            lambda_event_utils.inject_trigger_tags_for_sns(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.SQS:
            lambda_event_utils.inject_trigger_tags_for_sqs(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.S3:
            lambda_event_utils.inject_trigger_tags_for_s3(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.CloudWatchSchedule:
            lambda_event_utils.inject_trigger_tags_for_cloudwatch_schedule(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.CloudWatchLogs:
            lambda_event_utils.inject_trigger_tags_for_cloudwatch_logs(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.CloudFront:
            lambda_event_utils.inject_trigger_tags_for_cloudfront(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.APIGatewayProxy:
            lambda_event_utils.inject_trigger_tags_for_api_gateway_proxy(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.APIGateway:
            lambda_event_utils.inject_trigger_tags_for_api_gateway(span, original_event)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.Lambda:
            lambda_event_utils.inject_trigger_tags_for_lambda(span, original_context)
        elif lambda_event_type == lambda_event_utils.LambdaEventType.EventBridge:
            lambda_event_utils.inject_trigger_tags_for_eventbridge(span, original_event)
    except Exception as e:
        debug_logger("Cannot inject trigger tags. " + str(e))


def finish_trace(execution_context):
    root_span = execution_context.root_span
    scope = execution_context.scope
    try:
        root_span.finish(f_time=execution_context.finish_timestamp)
    except Exception:
        # TODO: handle root span finish errors
        pass
    finally:
        scope.close()

    trigger_class_name = root_span.get_tag(constants.SpanTags['TRIGGER_CLASS_NAME'])

    # Disable request data sending for cloudwatchlog, firehose and kinesis if not
    # enabled by configuration because requests can get too big for these
    enable_request_data = True
    if (
            trigger_class_name == constants.ClassNames['CLOUDWATCHLOG'] and
            not ConfigProvider.get(config_names.THUNDRA_LAMBDA_TRACE_CLOUDWATCHLOG_REQUEST_ENABLE)) or (

            trigger_class_name == constants.ClassNames['FIREHOSE'] and
            not ConfigProvider.get(config_names.THUNDRA_LAMBDA_TRACE_FIREHOSE_REQUEST_ENABLE)) or (

            trigger_class_name == constants.ClassNames['KINESIS'] and
            not ConfigProvider.get(config_names.THUNDRA_LAMBDA_TRACE_KINESIS_REQUEST_ENABLE)
    ):
        enable_request_data = False

    # ADDING TAGS #
    if (not ConfigProvider.get(config_names.THUNDRA_TRACE_REQUEST_SKIP)) and enable_request_data:
        root_span.set_tag('aws.lambda.invocation.request', execution_context.platform_data['originalEvent'])
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_RESPONSE_SKIP):
        root_span.set_tag('aws.lambda.invocation.response', execution_context.response)
        
    original_event = execution_context.platform_data["originalEvent"]
    if not check_lambda_authorizer(original_event) and trigger_class_name == constants.ClassNames['APIGATEWAY']:
        process_api_gw_response(execution_context)


def check_lambda_authorizer(original_event):
    if "type" not in original_event:
        return False
    elif original_event["type"] in {"REQUEST", "TOKEN", "COGNITO_USER_POOLS"}:
        return True
    return False

def process_api_gw_response(execution_context):
    try:
        if execution_context.response:
            response = execution_context.response
            if not response.get('headers'):
                response['headers'] = {}
            resource_path = utils.extract_api_gw_resource_name(execution_context.platform_data['originalEvent'])
            if resource_path:
                response['headers'][constants.TRIGGER_RESOURCE_NAME_KEY] = resource_path
    except:
        pass


def start_invocation(plugin_context, execution_context):
    execution_context.invocation_data = wrapper_utils.create_invocation_data(plugin_context, execution_context)
    execution_context.invocation_data['applicationPlatform'] = constants.CONTEXT_APPLICATION_PLATFORM


def get_response_status(execution_context):
    try:
        status_code = execution_context.response['statusCode']
    except:
        return None
    return status_code


def inject_step_function_info(execution_context, outgoing_trace_links):
    try:
        response = execution_context.response
        event = execution_context.platform_data['originalEvent']
        if ConfigProvider.get(config_names.THUNDRA_LAMBDA_AWS_STEPFUNCTIONS):
            trace_link = str(uuid.uuid4())
            step = 0
            if '_thundra' in event:
                step = event['_thundra']['step']

            if isinstance(response, dict):
                response['_thundra'] = {
                    'trace_link': trace_link,
                    'step': step + 1
                }
            outgoing_trace_links["outgoingTraceLinks"].append(trace_link)
    except Exception as e:
        print(e)

def inject_appsync_function_info(execution_context, outgoing_trace_links):
    try:
        response = execution_context.response
        if ConfigProvider.get(config_names.THUNDRA_LAMBDA_AWS_APPSYNC):
            trace_id = execution_context.trace_id

            if isinstance(response, dict):
                response['_thundra'] = {
                    'trace_id': trace_id,
                }
    except Exception as e:
        print(e)


def finish_invocation(execution_context):
    wrapper_utils.finish_invocation(execution_context)
    invocation_data = execution_context.invocation_data

    # Set response status code
    wrapper_utils.set_response_status(execution_context, get_response_status(execution_context))

    context = execution_context.platform_data['originalContext']

    _, used_mem = utils.process_memory_usage()
    used_mem_in_mb = used_mem / 1048576

    # Get outgoing trace links
    outgoing_trace_links = invocation_trace_support.get_outgoing_trace_links()

    # Inject trace link to response and add it to outgoing trace links for STEPFUNCTIONS
    inject_step_function_info(execution_context, outgoing_trace_links)
    # Inject trace link to response and add it to outgoing trace links for APPSYNC
    inject_appsync_function_info(execution_context, outgoing_trace_links)
    invocation_data.update(outgoing_trace_links)

    invocation_data['timeout'] = execution_context.timeout

    arn = getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None)

    # Add AWS tags
    invocation_data['tags']['aws.region'] = utils.get_aws_region_from_arn(
        getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, None))
    invocation_data['tags']['aws.lambda.name'] = getattr(context, constants.CONTEXT_FUNCTION_NAME, None)
    invocation_data['tags']['aws.lambda.arn'] = arn
    invocation_data['tags']['aws.account_no'] = utils.get_aws_account_no(arn)
    invocation_data['tags']['aws.lambda.memory_limit'] = int(getattr(context, constants.CONTEXT_MEMORY_LIMIT_IN_MB, 0))
    invocation_data['tags']['aws.lambda.log_group_name'] = getattr(context, constants.CONTEXT_LOG_GROUP_NAME, None)
    invocation_data['tags']['aws.lambda.log_stream_name'] = getattr(context, constants.CONTEXT_LOG_STREAM_NAME, None)
    invocation_data['tags']['aws.lambda.invocation.coldstart'] = invocation_data['coldStart']
    invocation_data['tags']['aws.lambda.invocation.timeout'] = execution_context.timeout
    invocation_data['tags']['aws.lambda.invocation.request_id'] = getattr(context, constants.CONTEXT_AWS_REQUEST_ID,
                                                                          None)
    invocation_data['tags']['aws.lambda.invocation.memory_usage'] = math.floor(used_mem_in_mb)

    xray_info = utils.parse_x_ray_trace_info()
    if xray_info.get("trace_id"):
        invocation_data['tags']['aws.xray.trace.id'] = xray_info.get("trace_id")
    if xray_info.get("segment_id"):
        invocation_data['tags']['aws.xray.segment.id'] = xray_info.get("segment_id")

    execution_context.invocation_data = invocation_data



import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.sqlalchemy import SqlAlchemyIntegration


def _wrapper(wrapped, instance, args, kwargs):
    engine = wrapped(*args, **kwargs)
    SqlAlchemyIntegration(engine)
    return engine


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_SQLALCHEMY_DISABLE):
        try:
            from sqlalchemy.event import listen
            from sqlalchemy.engine.interfaces import ExecutionContext
            wrapt.wrap_function_wrapper(
                'sqlalchemy',
                'create_engine',
                _wrapper
            )

            wrapt.wrap_function_wrapper(
                'sqlalchemy.engine',
                'create_engine',
                _wrapper
            )
        except:
            pass



import copy

import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.postgre import PostgreIntegration

postgre_integration = PostgreIntegration()


class PostgreCursorWrapper(wrapt.ObjectProxy):

    def __init__(self, cursor, connection_wrapper):
        super(PostgreCursorWrapper, self).__init__(cursor)
        self._self_connection = connection_wrapper

    def execute(self, *args, **kwargs):
        return postgre_integration.run_and_trace(
            self.__wrapped__.execute,
            self._self_connection,
            args,
            kwargs,
        )

    def callproc(self, *args, **kwargs):
        return postgre_integration.run_and_trace(
            self.__wrapped__.callproc,
            self._self_connection,
            args,
            kwargs,
        )

    def __enter__(self):
        # raise appropriate error if api not supported (should reach the user)
        value = self.__wrapped__.__enter__()
        if value is not self.__wrapped__:
            return value
        return self


class PostgreConnectionWrapper(wrapt.ObjectProxy):
    def cursor(self, *args, **kwargs):
        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return PostgreCursorWrapper(cursor, self)


def _wrapper(wrapped, instance, args, kwargs):
    connection = wrapped(*args, **kwargs)
    return PostgreConnectionWrapper(connection)


def _wrapper_register_type(wrapped, instance, args, kwargs):
    _args = list(copy.copy(args))
    if len(_args) == 2 and isinstance(_args[1], (PostgreConnectionWrapper, PostgreCursorWrapper)):
        _args[1] = _args[1].__wrapped__

    return wrapped(*_args, **kwargs)


def patch_extensions():
    _extensions = [
        ('psycopg2.extensions', 'register_type', _wrapper_register_type),
        ('psycopg2.extensions', 'quote_ident', _wrapper_register_type),
        ('psycopg2._psycopg', 'register_type', _wrapper_register_type)
    ]

    try:
        import psycopg2
        if getattr(psycopg2, '_json', None):
            _extensions.append(('psycopg2._json', 'register_type', _wrapper_register_type))
    except ImportError:
        pass

    for ext in _extensions:
        wrapt.wrap_function_wrapper(ext[0], ext[1], ext[2])


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_DISABLE):
        patch_extensions()
        wrapt.wrap_function_wrapper(
            'psycopg2',
            'connect',
            _wrapper)



import uuid

from thundra import constants, utils
from thundra.application.application_info_provider import ApplicationInfoProvider


class LambdaApplicationInfoProvider(ApplicationInfoProvider):

    def __init__(self):
        log_stream_name = utils.get_env_variable(constants.AWS_LAMBDA_LOG_STREAM_NAME)
        function_version = utils.get_env_variable(constants.AWS_LAMBDA_FUNCTION_VERSION)
        function_name = utils.get_env_variable(constants.AWS_LAMBDA_FUNCTION_NAME)
        region = utils.get_env_variable(constants.AWS_REGION, default='')

        application_instance_id = str(uuid.uuid4())
        if log_stream_name and len(log_stream_name.split(']')) >= 2:
            application_instance_id = log_stream_name.split(']')[1]

        self.application_info = {
            'applicationId': '',
            'applicationInstanceId': application_instance_id,
            'applicationName': function_name,
            'applicationVersion': function_version,
            'applicationRegion': region
        }

    def get_application_info(self):
        return self.application_info

    def get_application_tags(self):
        return self.application_info.get('applicationTags', {}).copy()

    @staticmethod
    def get_application_id(context, application_name=None):
        arn = getattr(context, constants.CONTEXT_INVOKED_FUNCTION_ARN, '')
        region = utils.get_aws_region_from_arn(arn)
        if not region:
            region = 'local'
        account_no = 'sam_local' if utils.sam_local_debugging() else utils.get_aws_account_no(arn)
        function_name = application_name if application_name else utils.get_aws_function_name(arn)
        application_id_template = 'aws:lambda:{region}:{account_no}:{function_name}'

        return application_id_template.format(region=region, account_no=account_no, function_name=function_name)

    def update(self, opts):
        filtered_opts = {k: v for k, v in opts.items() if v is not None}
        self.application_info.update(filtered_opts)



from thundra.samplers.base_sampler import BaseSampler

default_operator = 'or'
available_operators = ['and', 'or']


class CompositeSampler(BaseSampler):

    def __init__(self, samplers=None, operator=default_operator):
        if samplers is None:
            samplers = []
        self.samplers = samplers
        if operator in available_operators:
            self.operator = operator
        else:
            self.operator = default_operator

    def is_sampled(self, args=None):
        if len(self.samplers) == 0:
            return False

        if self.operator == 'or':
            sampled = False
            for sampler in self.samplers:
                sampled = sampler.is_sampled(args) or sampled
            return sampled
        elif self.operator == 'and':
            sampled = True
            for sampler in self.samplers:
                sampled = sampler.is_sampled(args) and sampled
            return sampled



from uuid import uuid4
from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils
import os, logging
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class AzureEnvironmentInfoProvider:

    ENVIRONMENT = "Azure"
    BUILD_BUILDID_ENV_VAR_NAME = "BUILD_BUILDID"
    BUILD_REPOSITORY_NAME_ENV_VAR_NAME = "BUILD_REPOSITORY_NAME"
    BUILD_REPOSITORY_URI_ENV_VAR_NAME = "BUILD_REPOSITORY_URI"
    BUILD_SOURCEBRANCHNAME_ENV_VAR_NAME = "BUILD_SOURCEBRANCHNAME"
    BUILD_SOURCEVERSION_ENV_VAR_NAME = "BUILD_SOURCEVERSION"
    BUILD_SOURCEVERSIONMESSAGE_ENV_VAR_NAME = "BUILD_SOURCEVERSIONMESSAGE"
    INVOCATION_ID_ENV_VAR_NAME = "INVOCATION_ID"
    BUILD_REPOSITORY_PROVIDER_ENV_VAR_NAME = "BUILD_REPOSITORY_PROVIDER"

    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash, environment):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        build_number = os.getenv(cls.BUILD_BUILDID_ENV_VAR_NAME)
        invocation_id = os.getenv(cls.INVOCATION_ID_ENV_VAR_NAME)
        test_run_key = cls.generate_test_run_key(invocation_id, build_number)
        if test_run_key:
            return TestRunnerUtils.get_test_run_id(environment, repo_url, commit_hash, test_run_key)
        else:
            return TestRunnerUtils.get_default_test_run_id(environment, repo_url, commit_hash)


    @staticmethod
    def generate_test_run_key(invocation_id=None, build_number=None):
        if invocation_id and build_number:
            return invocation_id + "-" + build_number
        elif invocation_id:
            return invocation_id
        elif build_number:
            return build_number
        else:
            return None

    @classmethod
    def build_env_info(cls):
        try:
            repo_url = os.getenv(cls.BUILD_REPOSITORY_URI_ENV_VAR_NAME)
            repo_name = os.getenv(cls.BUILD_REPOSITORY_NAME_ENV_VAR_NAME) or GitHelper.extractRepoName(repo_url)
            branch = os.getenv(cls.BUILD_SOURCEBRANCHNAME_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.BUILD_SOURCEVERSION_ENV_VAR_NAME)
            commit_message = os.getenv(cls.BUILD_SOURCEVERSIONMESSAGE_ENV_VAR_NAME) or GitHelper.get_commit_message()
            repo_provider = os.getenv(cls.BUILD_REPOSITORY_PROVIDER_ENV_VAR_NAME)
            
            environment = cls.ENVIRONMENT + " - " + str(repo_provider) if repo_provider else cls.ENVIRONMENT

            if not branch:
                branch = GitHelper.get_branch()

            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()

            test_run_id = cls.get_test_run_id(repo_url, commit_hash, environment)

            env_info = EnvironmentInfo(test_run_id, environment, repo_url, repo_name, 
                branch, commit_hash, commit_message)
            print_debug_message_to_console("Azure Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("Unable to build environment info: {}".format(err))
            LOGGER.error("Unable to build environment info: {}".format(err))
            pass
        return None


from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils 
import os, logging
from foresight.utils.generic_utils import print_debug_message_to_console


LOGGER = logging.getLogger(__name__)


class TravisCIEnvironmentInfoProvider:
    ENVIRONMENT = "TravisCI"
    TRAVIS_REPO_SLUG_VAR_NAME = "TRAVIS_REPO_SLUG"
    TRAVIS_PULL_REQUEST_BRANCH_ENV_VAR_NAME = "TRAVIS_PULL_REQUEST_BRANCH"
    TRAVIS_BRANCH_ENV_VAR_NAME = "TRAVIS_BRANCH"
    TRAVIS_COMMIT_ENV_VAR_NAME = "TRAVIS_COMMIT"
    TRAVIS_COMMIT_MESSAGE_ENV_VAR_NAME = "TRAVIS_COMMIT_MESSAGE"
    TRAVIS_BUILD_WEB_URL_ENV_VAR_NAME = "TRAVIS_BUILD_WEB_URL"
    TRAVIS_BUILD_ID_ENV_VAR_NAME = "TRAVIS_BUILD_ID"


    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        build_web_url = os.getenv(cls.TRAVIS_BUILD_WEB_URL_ENV_VAR_NAME)
        build_id = os.getenv(cls.TRAVIS_BUILD_ID_ENV_VAR_NAME)
        test_run_key = TestRunnerUtils.string_concat_by_underscore(build_web_url, build_id)
        if test_run_key != "":
            return TestRunnerUtils.get_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash, 
                test_run_key)
        else:
            return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)


    @classmethod
    def build_env_info(cls):
        try:
            repo_url = "https://github.com/{}.git".format(os.getenv(cls.TRAVIS_REPO_SLUG_VAR_NAME))
            repo_name = GitHelper.extractRepoName(repo_url)
            branch = os.getenv(cls.TRAVIS_PULL_REQUEST_BRANCH_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.TRAVIS_COMMIT_ENV_VAR_NAME)
            commit_message = os.getenv(cls.TRAVIS_COMMIT_MESSAGE_ENV_VAR_NAME)

            if not branch:
                branch = GitHelper.get_branch()
            
            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()

            if not commit_message:
                commit_message = GitHelper.get_commit_message()

            test_run_id = cls.get_test_run_id(repo_url, commit_hash)
            
            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, 
                repo_name, branch, commit_hash, commit_message)
            print_debug_message_to_console("TravisCI Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("TravisCI Unable to build environment info: {}".format(err))
            LOGGER.error("TravisCI Unable to build environment info: {}".format(err))
            pass
        return None


import uuid
from threading import Lock

import opentracing
from opentracing import Format

from thundra import constants
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.opentracing.propagation.http import HTTPPropagator
from thundra.opentracing.propagation.text import TextMapPropagator
from thundra.opentracing.span import ThundraSpan
from thundra.opentracing.span_context import ThundraSpanContext
from thundra.plugins.trace import trace_support

from thundra.utils import arrange_scope_manager


class ThundraTracer(opentracing.Tracer):
    __instance = None
    

    @staticmethod
    def get_instance():
        return ThundraTracer() if ThundraTracer.__instance is None else ThundraTracer.__instance


    def __init__(self, scope_manager=None):
        scope_manager = arrange_scope_manager(scope_manager)
        super(ThundraTracer, self).__init__(scope_manager)
        self.lock = Lock()
        self.global_span_order = 0
        ThundraTracer.__instance = self
        self._propagators = {
            Format.HTTP_HEADERS: HTTPPropagator(),
            Format.TEXT_MAP: TextMapPropagator(),
        }

    def start_active_span(self,
                          operation_name,
                          child_of=None,
                          references=None,
                          trace_id=None,
                          transaction_id=None,
                          span_id=None,
                          parent_span_id=None,
                          tags=None,
                          start_time=None,
                          span_order=-1,
                          ignore_active_span=False,
                          finish_on_close=True,
                          execution_context=None):
        span_id = span_id or str(uuid.uuid4())
        _span = self.start_span(operation_name=operation_name,
                                child_of=child_of,
                                references=references,
                                trace_id=trace_id,
                                transaction_id=transaction_id,
                                span_id=span_id,
                                parent_span_id=parent_span_id,
                                tags=tags,
                                start_time=start_time,
                                span_order=span_order,
                                ignore_active_span=ignore_active_span,
                                execution_context=execution_context)
        return self.scope_manager.activate(_span, finish_on_close)

    def start_span(self,
                   operation_name=None,
                   class_name=None,
                   domain_name=None,
                   child_of=None,
                   references=None,
                   trace_id=None,
                   transaction_id=None,
                   span_id=None,
                   parent_span_id=None,
                   tags=None,
                   start_time=None,
                   span_order=-1,
                   ignore_active_span=False,
                   execution_context=None):
        # Create a new span
        _span = self.create_span(operation_name=operation_name,
                                 class_name=class_name,
                                 domain_name=domain_name,
                                 child_of=child_of,
                                 references=references,
                                 trace_id=trace_id,
                                 transaction_id=transaction_id,
                                 span_id=span_id,
                                 parent_span_id=parent_span_id,
                                 tags=tags,
                                 start_time=start_time,
                                 span_order=span_order,
                                 ignore_active_span=ignore_active_span,
                                 execution_context=execution_context)

        # Record the new span
        if hasattr(_span, 'execution_context') and _span.execution_context:
            recorder = _span.execution_context.recorder
            recorder.record(_span)
        # Update parent span tag for line by line tracing
        self.inject_line_by_line_tags(_span)

        # Call listener's on_span_started method with the new span
        return _span

    def create_span(self,
                    operation_name=None,
                    class_name=None,
                    domain_name=None,
                    child_of=None,
                    references=None,
                    trace_id=None,
                    transaction_id=None,
                    span_id=None,
                    parent_span_id=None,
                    tags=None,
                    start_time=None,
                    span_order=-1,
                    ignore_active_span=False,
                    execution_context=None):

        with self.lock:
            self.global_span_order += 1

        _span_order = span_order
        if _span_order == -1:
            _span_order = self.global_span_order

        _parent_context = None
        if child_of is not None:
            _parent_context = child_of if isinstance(child_of, opentracing.SpanContext) else child_of.context
            if isinstance(child_of, opentracing.Span) and hasattr(child_of,
                                                                  'execution_context') and child_of.execution_context:
                execution_context = child_of.execution_context
        elif references is not None and len(references) > 0:
            _parent_context = references[0].referenced_context

        _scope = self.scope_manager.active
        if _scope is not None and _scope.span is not None and execution_context is None:
            if hasattr(_scope.span, 'execution_context') and _scope.span.execution_context:
                execution_context = _scope.span.execution_context

        if not ignore_active_span and _parent_context is None:
            if _scope is not None and _scope.span is not None:
                _parent_context = _scope.span.context

        _trace_id = trace_id
        _transaction_id = transaction_id
        _span_id = span_id or str(uuid.uuid4())
        _parent_span_id = parent_span_id

        if _parent_context is not None:
            _trace_id = _trace_id or _parent_context.trace_id
            _transaction_id = _transaction_id or _parent_context.transaction_id
            _parent_span_id = _parent_span_id or _parent_context.span_id

        _context = ThundraSpanContext(trace_id=_trace_id,
                                      transaction_id=_transaction_id,
                                      span_id=_span_id,
                                      parent_span_id=_parent_span_id)
        _span = ThundraSpan(self,
                            operation_name=operation_name,
                            class_name=class_name,
                            domain_name=domain_name,
                            context=_context,
                            tags=tags,
                            start_time=start_time,
                            span_order=_span_order,
                            execution_context=execution_context)

        return _span

    def get_active_span(self):
        scope = self.scope_manager.active
        if scope is not None:
            return scope.span

        return None

    def get_span_listeners(self):
        return trace_support.get_span_listeners()

    def get_spans(self):
        execution_context = ExecutionContextManager.get()
        if execution_context and execution_context.recorder:
            return execution_context.recorder.get_spans()
        return []

    def clear(self):
        execution_context = ExecutionContextManager.get()
        if execution_context and execution_context.recorder:
            execution_context.recorder.clear()

    def inject(self, span_context, format, carrier):
        propagator = self._propagators.get(format, None)

        if propagator is None:
            raise opentracing.UnsupportedFormatException

        propagator.inject(span_context, carrier)

    def extract(self, format, carrier):
        propagator = self._propagators.get(format, None)

        if propagator is None:
            raise opentracing.UnsupportedFormatException

        return propagator.extract(carrier)

    def add_span_listener(self, listener):
        trace_support.register_span_listener(listener)

    def inject_line_by_line_tags(self, span):
        parent_span = self.get_active_span()

        try:
            if parent_span and parent_span.get_tag(constants.LineByLineTracingTags['lines']):
                lines = parent_span.get_tag(constants.LineByLineTracingTags['lines'])
                last_line = lines[-1]
                next_span_ids = last_line.get(constants.LineByLineTracingTags['next_span_ids'])
                if not next_span_ids:
                    next_span_ids = []

                next_span_ids.append(span.span_id)
                last_line[constants.LineByLineTracingTags['next_span_ids']] = next_span_ids
                parent_span.set_tag(constants.LineByLineTracingTags['lines'], lines)
        except Exception:
            pass



class EnvironmentInfo:

    def __init__(self, test_run_id, environment=None, repo_url=None, repo_name=None, 
        branch=None, commit_hash=None, commit_message=None):
        
        self.test_run_id = test_run_id
        self.environment = environment
        self.repo_url = repo_url
        self.repo_name = repo_name
        self.branch = branch
        self.commit_hash = commit_hash
        self.commit_message = commit_message

    def to_json(self):
        return {
            "testRunId": self.test_run_id,
            "environment": self.environment,
            "repoURL": self.repo_url,
            "repoName": self.repo_name,
            "branch": self.branch,
            "commitHash": self.commit_hash,
            "commitMessage": self.commit_message
        }


from thundra.samplers.base_sampler import BaseSampler


class DurationAwareSampler(BaseSampler):

    def __init__(self, duration, longer_than=False):
        self.duration = duration
        self.longer_than = longer_than

    def is_sampled(self, span=None):
        if not span:
            return False

        if self.longer_than:
            return span.get_duration() > self.duration

        return span.get_duration() <= self.duration



import wrapt

from thundra import utils
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.requests import RequestsIntegration

request_integration = RequestsIntegration()


def _wrapper(wrapped, instance, args, kwargs):
    prepared_request = args[0]

    if utils.is_excluded_url(prepared_request.url):
        return wrapped(*args, **kwargs)

    return request_integration.run_and_trace(
        wrapped,
        instance,
        args,
        kwargs,
    )


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_DISABLE):
        wrapt.wrap_function_wrapper(
            'requests',
            'Session.send',
            _wrapper
        )



import gc
import logging
import threading
import time
import uuid

from thundra import utils, constants
from thundra.compat import PY2
from thundra.opentracing.tracer import ThundraTracer
from thundra.plugins.config.metric_config import MetricConfig
from thundra.plugins.metric import metric_support

logger = logging.getLogger(__name__)


class MetricPlugin:

    def __init__(self, plugin_context=None, config=None):
        self.metric_data = {}
        self.plugin_context = plugin_context
        self.tracer = ThundraTracer.get_instance()
        self.system_cpu_usage_start = float(0)
        self.system_cpu_usage_end = float(0)
        self.system_cpu_total_start = float(0)
        self.system_cpu_total_end = float(0)
        self.process_cpu_usage_start = float(0)
        self.process_cpu_usage_end = float(0)
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }

        if isinstance(config, MetricConfig):
            self.config = config
        else:
            self.config = MetricConfig()

    def before_invocation(self, execution_context):
        metric_time = time.time() * 1000
        active_span = self.tracer.get_active_span()

        self.metric_data = {
            'type': "Metric",
            'agentVersion': constants.THUNDRA_AGENT_VERSION,
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'traceId': active_span.trace_id if active_span is not None else '',
            'transactionId': execution_context.transaction_id,
            'spanId': active_span.span_id if active_span is not None else '',
            'metricTimestamp': int(metric_time),
            'tags': {
                'aws.region': utils.get_env_variable(constants.AWS_REGION, default='')
            }
        }
        # Add application related data
        application_info = self.plugin_context.application_info
        self.metric_data.update(application_info)
        self.system_cpu_total_start, self.system_cpu_usage_start = utils.system_cpu_usage()
        self.process_cpu_usage_start = utils.process_cpu_usage()

    def after_invocation(self, execution_context):
        if not self.check_sampled():
            return

        self.add_thread_metric_report(execution_context)
        self.add_gc_metric_report(execution_context)
        self.add_memory_metric_report(execution_context)
        self.add_cpu_metric_report(execution_context)

    def check_sampled(self):
        sampler = metric_support.get_sampler()
        if self.config.sampler:
            sampler = self.config.sampler
        sampled = True
        if sampler is not None:
            try:
                sampled = sampler.is_sampled(self.metric_data)
            except Exception as e:
                logger.error("error while sampling metrics: %s", e)
        return sampled

    def add_thread_metric_report(self, execution_context):
        active_thread_counts = threading.active_count()
        thread_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'ThreadMetric',
        }
        metrics = {
            'threadCount': active_thread_counts if active_thread_counts is not None else -1
        }
        thread_metric_data.update(self.metric_data)
        thread_metric_data['metrics'] = metrics
        thread_metric_report = {
            'data': thread_metric_data,
            'type': 'Metric',
            'apiKey': self.plugin_context.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        execution_context.report(thread_metric_report)

    def add_gc_metric_report(self, execution_context):
        gc_metrics = gc.get_count()
        if not PY2:
            gc_metrics = gc.get_stats()

        gc_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'GCMetric'
        }

        gen = 0
        metrics = {}
        for metric in gc_metrics:
            key = 'generation' + str(gen) + 'Collections'
            if PY2:
                metrics[key] = metric
            else:
                metrics[key] = metric['collections']
            gen += 1
        gc_metric_data.update(self.metric_data)
        gc_metric_data['metrics'] = metrics
        gc_metric_report = {
            'data': gc_metric_data,
            'type': 'Metric',
            'apiKey': self.plugin_context.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        execution_context.report(gc_metric_report)

    def add_memory_metric_report(self, execution_context):
        size, used = utils.process_memory_usage()
        memory_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'MemoryMetric',
        }
        metrics = {
            'app.maxMemory': size,
            'app.usedMemory': used
        }
        memory_metric_data.update(self.metric_data)
        memory_metric_data['metrics'] = metrics
        memory_metric_report = {
            'data': memory_metric_data,
            'type': 'Metric',
            'apiKey': self.plugin_context.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        execution_context.report(memory_metric_report)

    def add_cpu_metric_report(self, execution_context):
        self.process_cpu_usage_end = utils.process_cpu_usage()
        self.system_cpu_total_end, self.system_cpu_usage_end = utils.system_cpu_usage()
        process_cpu_load = 0
        system_cpu_load = 0
        system_cpu_total = self.system_cpu_total_end - self.system_cpu_total_start
        system_cpu_usage = self.system_cpu_usage_end - self.system_cpu_usage_start
        process_cpu_usage = self.process_cpu_usage_end - self.process_cpu_usage_start
        if system_cpu_total != 0:
            cpu_load = process_cpu_usage / system_cpu_total
            process_cpu_load = 1 if cpu_load > 1 else cpu_load
            cpu_load = system_cpu_usage / system_cpu_total
            system_cpu_load = 1 if cpu_load > 1 else cpu_load

        cpu_metric_data = {
            'id': str(uuid.uuid4()),
            'metricName': 'CPUMetric'
        }

        metrics = {
            'app.cpuLoad': process_cpu_load if process_cpu_load is not None else -1,
            'sys.cpuLoad': system_cpu_load if system_cpu_load is not None else -1
        }
        cpu_metric_data.update(self.metric_data)
        cpu_metric_data['metrics'] = metrics
        cpu_metric_report = {
            'data': cpu_metric_data,
            'type': 'Metric',
            'apiKey': self.plugin_context.api_key,
            'dataModelVersion': constants.DATA_FORMAT_VERSION
        }
        execution_context.report(cpu_metric_report)



import traceback
import time
import logging
from thundra import constants
from thundra.opentracing.tracer import ThundraTracer
from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names


try:
    from pymongo.monitoring import CommandListener
    from bson.json_util import dumps
except ImportError:
    CommandListener = object


logger = logging.getLogger(__name__)


class CommandTracer(CommandListener):
    _scopes = {}

    def started(self, event):
        tracer = ThundraTracer.get_instance()
        if not tracer.get_active_span():
            return

        scope = tracer.start_active_span(operation_name=event.database_name, finish_on_close=False)

        self._scopes[event.request_id] = scope

        # Inject before span tags
        try:
            command_name = event.command_name
            scope.span.class_name = constants.ClassNames['MONGODB']
            scope.span.domain_name = constants.DomainNames['DB']

            try:
                host, port = event.connection_id
            except:
                host, port = "", ""

            try:
                collection_name = next(iter(event.command.items()))[1]
            except:
                collection_name = ""

            operation_type = constants.MongoDBCommandTypes.get(command_name.upper(), '')

            tags = {
                constants.SpanTags['OPERATION_TYPE']: operation_type,
                constants.DBTags['DB_TYPE']: 'mongodb',
                constants.DBTags['DB_HOST']: host,
                constants.DBTags['DB_PORT']: port,
                constants.SpanTags['DB_INSTANCE']: event.database_name,
                constants.MongoDBTags['MONGODB_COMMAND_NAME']: command_name.upper(),
                constants.MongoDBTags['MONGODB_COLLECTION']: collection_name,
                constants.SpanTags['TOPOLOGY_VERTEX']: True,
            }

            if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_MONGODB_COMMAND_MASK):
                try:
                    tags[constants.MongoDBTags['MONGODB_COMMAND']] = dumps(event.command)[
                                                                   :constants.DEFAULT_MONGO_COMMAND_SIZE_LIMIT]
                except Exception as e:
                    tags[constants.MongoDBTags['MONGODB_COMMAND']] = ''

            scope.span.tags = tags

        except Exception as instrumentation_exception:
            e = {
                'type': str(type(instrumentation_exception)),
                'message': str(instrumentation_exception),
                'traceback': traceback.format_exc(),
                'time': time.time()
            }
            scope.span.set_tag('instrumentation_error', e)

        try:
            # Inform span that initialization completed
            scope.span.on_started()
        except Exception as e:
            logger.error(e)

    def succeeded(self, event):
        scope = self._scopes.pop(event.request_id, None)
        if scope is None:
            return
        try:
            scope.span.finish()
        except Exception as e:
            logger.error(e)

        scope.close()

    def failed(self, event):
        scope = self._scopes.pop(event.request_id, None)
        if scope is None:
            return

        scope.span.set_error_to_tag(event.failure)
        try:
            scope.span.finish()
        except Exception as e:
            logger.error(e)

        scope.close()



import logging
import traceback
import uuid

from thundra import constants
from thundra.opentracing.tracer import ThundraTracer
from thundra.plugins.config.trace_config import TraceConfig
from thundra.plugins.invocation import invocation_support
from thundra.plugins.trace import trace_support

logger = logging.getLogger(__name__)


class TracePlugin:

    def __init__(self, plugin_context, config=None):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.tracer = ThundraTracer.get_instance()
        self.plugin_context = plugin_context
        if isinstance(config, TraceConfig):
            self.config = config
        else:
            self.config = TraceConfig()

    def before_invocation(self, execution_context):
        executor = self.plugin_context.executor
        if executor:
            executor.start_trace(self.plugin_context, execution_context, self.tracer)

    def after_invocation(self, execution_context):
        executor = self.plugin_context.executor
        if executor:
            executor.finish_trace(execution_context)

        span_stack = execution_context.recorder.get_spans()

        sampled = True
        if len(span_stack) > 0:
            sampled = self.check_sampled(span_stack[0])

        span_data_list = []
        for span in span_stack:
            if sampled:
                current_span_data = self.wrap_span(self.build_span(span, execution_context),
                                                   self.plugin_context.api_key)
                span_data_list.append(current_span_data)

        execution_context.recorder.clear()

        if execution_context.error:
            error = execution_context.error
            self.set_error_to_root_span(execution_context.root_span, error)
        elif execution_context.user_error:
            self.set_error_to_root_span(execution_context.root_span, execution_context.user_error)

        execution_context.report(span_data_list)

        invocation_support.clear_error()

    @staticmethod
    def set_error_to_root_span(root_span, error):
        root_span.set_tag('error', True)
        if isinstance(error, Exception):
            error_type = type(error)
            root_span.set_tag('error.kind', error_type.__name__)
            root_span.set_tag('error.message', str(error))

            if hasattr(error, 'code'):
                root_span.set_tag('error.code', error.code)
            if hasattr(error, '__traceback__'):
                root_span.set_tag('error.stack', ''.join(
                    traceback.format_exception(error_type, error, error.__traceback__)))

        elif isinstance(error, dict):
            root_span.set_tag('error.kind', error.get('type'))
            root_span.set_tag('error.message', error.get('message'))
            if error.get('traceback'):
                root_span.set_tag('error.stack', error.get('traceback'))

    def build_span(self, span, execution_context):
        if not execution_context.transaction_id:
            execution_context.transaction_id = str(uuid.uuid4())
        service_name = span.service_name if hasattr(span, "service_name") else ''
        span_data = {
            'id': span.context.span_id,
            'type': "Span",
            'agentVersion': constants.THUNDRA_AGENT_VERSION,
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'traceId': span.context.trace_id,
            'transactionId': execution_context.transaction_id,
            'parentSpanId': span.context.parent_span_id or '',
            'spanOrder': span.span_order,
            'domainName': span.domain_name or '',
            'className': span.class_name or '',
            'serviceName': service_name,
            'operationName': span.operation_name,
            'startTimestamp': span.start_time,
            'finishTimestamp': span.finish_time,
            'duration': span.get_duration(),
            'logs': span.logs,
            'tags': span.tags
        }

        # Add application related data
        application_info = self.plugin_context.application_info
        span_data.update(application_info)

        return span_data

    @staticmethod
    def wrap_span(span_data, api_key):
        report_data = {
            'apiKey': api_key,
            'type': 'Span',
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'data': span_data
        }

        return report_data

    def check_sampled(self, span):
        sampler = trace_support.get_sampler()
        if self.config.sampler:
            sampler = self.config.sampler

        sampled = True
        if sampler is not None:
            try:
                sampled = sampler.is_sampled(span)
            except Exception as e:
                logger.error("error while sampling spans: %s", e)
        return sampled



from thundra.context.context_provider import ContextProvider
from thundra.context.execution_context import ExecutionContext
from thundra.opentracing.tracer import ThundraTracer


class TracingExecutionContextProvider(ContextProvider):
    def __init__(self):
        self.tracer = ThundraTracer.get_instance()

    def get(self):
        execution_context = None
        active_span = self.tracer.get_active_span()
        if active_span and hasattr(active_span, 'execution_context'):
            execution_context = active_span.execution_context
        if not execution_context:
            return ExecutionContext()
        return execution_context
    
    def set(self, execution_context):
        active_span = self.tracer.get_active_span()
        if active_span and hasattr(active_span, 'execution_context'):
            active_span.execution_context = execution_context

    def clear(self):
        active_span = self.tracer.get_active_span()
        if active_span and hasattr(active_span, 'execution_context'):
            active_span.execution_context = None




from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.base_integration import BaseIntegration
from thundra.integrations.rdb_base import RdbBaseIntegration


class MysqlIntegration(BaseIntegration, RdbBaseIntegration):
    CLASS_TYPE = 'mysql'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        return instance.database

    def before_call(self, scope, cursor, connection, _args, _kwargs, response, exception):
        span = scope.span
        span.domain_name = constants.DomainNames['DB']
        span.class_name = constants.ClassNames['MYSQL']

        query = ''
        operation = ''
        try:
            query = _args[0]
            if len(query) > 0:
                operation = query.split()[0].strip("\"").lower()
        except Exception:
            pass

        tags = {
            constants.SpanTags['OPERATION_TYPE']: MysqlIntegration._OPERATION_TO_TYPE.get(operation, ''),
            constants.SpanTags['DB_INSTANCE']: connection._database,
            constants.SpanTags['DB_HOST']: connection._host,
            constants.SpanTags['DB_TYPE']: "mysql",
            constants.SpanTags['DB_STATEMENT_TYPE']: operation.upper(),
            constants.SpanTags['TOPOLOGY_VERTEX']: True,
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_RDB_STATEMENT_MASK):
            tags[constants.DBTags['DB_STATEMENT']] = query

        span.tags = tags



import abc
from thundra.constants import (TEST_RUN_EVENTS_DATA_VERSION,
     DATA_FORMAT_VERSION,
     THUNDRA_AGENT_VERSION)
from thundra.config.config_provider import ConfigProvider
from thundra.config.config_names import THUNDRA_APIKEY


ABC = abc.ABCMeta('ABC', (object,), {})

class TestRunMonitoring(ABC):

    TEST_RUN_DATA_MODEL_VERSION = TEST_RUN_EVENTS_DATA_VERSION
    DATA_FORMAT_VERSION = DATA_FORMAT_VERSION
    AGENT_VERSION = THUNDRA_AGENT_VERSION

    def get_monitoring_data(self):
        return {
            'apiKey': ConfigProvider.get(THUNDRA_APIKEY, None),
            'type': None,
            'dataModelVersion': self.DATA_FORMAT_VERSION,
            'data': None
        }


import signal
import sys
import threading


class ThreadTimeout(object):
    def __init__(self, seconds, handler, execution_context):
        self.seconds = seconds
        self.timer = None
        self.handler = handler
        self.execution_context = execution_context

    def __enter__(self):
        self.timer = threading.Timer(self.seconds, self.handler, [self.execution_context])
        if self.seconds > 0:
            self.timer.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.timer:
            self.timer.cancel()
        return False


class SignalTimeout(object):
    def __init__(self, seconds, handler, execution_context):
        self.seconds = seconds
        self.handler = handler
        self.execution_context = execution_context

    def __enter__(self):
        if threading.current_thread().__class__.__name__ == '_MainThread':
            signal.signal(signal.SIGALRM, self._timeout)
            if self.seconds > 0:
                signal.setitimer(signal.ITIMER_REAL, self.seconds)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if threading.current_thread().__class__.__name__ == '_MainThread':
            signal.setitimer(signal.ITIMER_REAL, 0)
        return False

    def _timeout(self, signum, frame):
        current_thread = threading.current_thread().__class__.__name__
        if current_thread == '_MainThread' and signum == signal.SIGALRM:
            self.handler(self.execution_context)


if sys.platform == "win32":
    Timeout = ThreadTimeout
else:
    Timeout = SignalTimeout



from thundra import constants

from thundra.opentracing.propagation.text import TextMapPropagator
from thundra.opentracing.span_context import ThundraSpanContext


class HTTPPropagator(TextMapPropagator):

    @staticmethod
    def extract_value_from_header(header_name, headers):
        for header, value in headers.items():
            if header.lower() == header_name.lower():
                return value

    def extract(self, carrier):
        try:
            trace_id = HTTPPropagator.extract_value_from_header(constants.THUNDRA_TRACE_ID_KEY, carrier)
            span_id = HTTPPropagator.extract_value_from_header(constants.THUNDRA_SPAN_ID_KEY, carrier)
            transaction_id = HTTPPropagator.extract_value_from_header(constants.THUNDRA_TRANSACTION_ID_KEY, carrier)
        except:
            return None

        if not (trace_id and span_id and transaction_id):
            return None

        # Extract baggage items
        baggage = {}
        for key in carrier:
            if isinstance(key, str):
                if key.startswith(constants.THUNDRA_BAGGAGE_PREFIX):
                    baggage[key[len(constants.THUNDRA_BAGGAGE_PREFIX):]] = carrier[key]
            elif isinstance(key, tuple):
                if key[0].startswith(constants.THUNDRA_BAGGAGE_PREFIX):
                    baggage[key[0][len(constants.THUNDRA_BAGGAGE_PREFIX):]] = key[1]

        span_context = ThundraSpanContext(trace_id=trace_id, span_id=span_id, transaction_id=transaction_id,
                                          baggage=baggage)

        return span_context



from thundra.application.application_info_provider import ApplicationInfoProvider
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider


def test_if_can_get_integer_tag():
    tag_name = 'integerField'
    (env_key, env_val) = (config_names.THUNDRA_APPLICATION_TAG_PREFIX + '.' + tag_name, 3773)
    ConfigProvider.set(env_key, str(env_val))

    application_tags = ApplicationInfoProvider.parse_application_tags()

    assert application_tags[tag_name] == env_val


def test_if_can_get_float_tag():
    tag_name = 'floatField'
    (env_key, env_val) = (config_names.THUNDRA_APPLICATION_TAG_PREFIX + '.' + tag_name, 12.3221)
    ConfigProvider.set(env_key, str(env_val))

    application_tags = ApplicationInfoProvider.parse_application_tags()

    assert application_tags[tag_name] == env_val


def test_if_can_get_string_tag():
    tag_name = 'stringField'
    (env_key, env_val) = (config_names.THUNDRA_APPLICATION_TAG_PREFIX + '.' + tag_name, 'fooBar')
    ConfigProvider.set(env_key, str(env_val))

    application_tags = ApplicationInfoProvider.parse_application_tags()

    assert application_tags[tag_name] == env_val


def test_if_can_get_bool_tag():
    tag_name = 'boolField'
    (env_key, env_val) = (config_names.THUNDRA_APPLICATION_TAG_PREFIX + '.' + tag_name, True)
    ConfigProvider.set(env_key, str(env_val))

    application_tags = ApplicationInfoProvider.parse_application_tags()

    assert application_tags[tag_name] == env_val



from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils
import logging
import os
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class GitlabEnvironmentInfoProvider:

    ENVIRONMENT = "GitLab"
    CI_REPOSITORY_URL_ENV_VAR_NAME = "CI_REPOSITORY_URL"
    CI_COMMIT_BRANCH_ENV_VAR_NAME = "CI_COMMIT_BRANCH"
    CI_COMMIT_REF_NAME_ENV_VAR_NAME = "CI_COMMIT_REF_NAME"
    CI_COMMIT_SHA_ENV_VAR_NAME = "CI_COMMIT_SHA"
    CI_COMMIT_MESSAGE_ENV_VAR_NAME = "CI_COMMIT_MESSAGE"
    CI_JOB_ID_ENV_VAR_NAME = "CI_JOB_ID"
    CI_JOB_URL_ENV_VAR_NAME = "CI_JOB_URL"


    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        job_id = os.getenv(cls.CI_JOB_ID_ENV_VAR_NAME)
        job_url = os.getenv(cls.CI_JOB_URL_ENV_VAR_NAME)
        test_run_key = TestRunnerUtils.string_concat_by_underscore(job_id, job_url)
        if test_run_key != "":
            return TestRunnerUtils.get_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash, 
                test_run_key)
        else:
            return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)


    @classmethod
    def build_env_info(cls):
        try:
            repo_url = os.getenv(cls.CI_REPOSITORY_URL_ENV_VAR_NAME)
            repo_name = GitHelper.extractRepoName(repo_url)
            branch = os.getenv(cls.CI_COMMIT_BRANCH_ENV_VAR_NAME)
            if not branch:
                branch = os.getenv(cls.CI_COMMIT_REF_NAME_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.CI_COMMIT_SHA_ENV_VAR_NAME)
            commit_message = os.getenv(cls.CI_COMMIT_MESSAGE_ENV_VAR_NAME)

            if not branch:
                branch = GitHelper.get_branch()
            
            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()

            if not commit_message:
                commit_message = GitHelper.get_commit_message()

            test_run_id = cls.get_test_run_id(repo_url, commit_hash)

            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, repo_name, branch, commit_hash, commit_message)
            print_debug_message_to_console("Gitlab Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("Gitlab Unable to build environment info: {}".format(err))
            LOGGER.error("Gitlab Unable to build environment info: {}".format(err))
            pass
        return None


from foresight.environment.git.git_helper import GitHelper
from foresight.environment.environment_info import EnvironmentInfo
from foresight.utils.test_runner_utils import TestRunnerUtils
import os, logging
from foresight.utils.generic_utils import print_debug_message_to_console

LOGGER = logging.getLogger(__name__)

class BitbucketEnvironmentInfoProvider:

    ENVIRONMENT = "BitBucket"
    BITBUCKET_GIT_HTTP_ORIGIN_ENV_VAR_NAME = "BITBUCKET_GIT_HTTP_ORIGIN"
    BITBUCKET_GIT_SSH_ORIGIN_ENV_VAR_NAME = "BITBUCKET_GIT_SSH_ORIGIN"
    BITBUCKET_BRANCH_ENV_VAR_NAME = "BITBUCKET_BRANCH"
    BITBUCKET_COMMIT_ENV_VAR_NAME = "BITBUCKET_COMMIT"
    BITBUCKET_BUILD_NUMBER_ENV_VAR_NAME = "BITBUCKET_BUILD_NUMBER"

    @classmethod
    def get_test_run_id(cls, repo_url, commit_hash):
        configured_test_run_id = TestRunnerUtils.get_configured_test_run_id()
        if configured_test_run_id:
            return configured_test_run_id
        build_number = os.getenv(cls.BITBUCKET_BUILD_NUMBER_ENV_VAR_NAME)
        if build_number:
            return TestRunnerUtils.get_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash, build_number)
        else:
            return TestRunnerUtils.get_default_test_run_id(cls.ENVIRONMENT, repo_url, commit_hash)


    @classmethod
    def build_env_info(cls):
        try:
            repo_url = os.getenv(cls.BITBUCKET_GIT_HTTP_ORIGIN_ENV_VAR_NAME)
            if not repo_url:
                repo_url = os.getenv(cls.BITBUCKET_GIT_SSH_ORIGIN_ENV_VAR_NAME)
            repo_name = GitHelper.extractRepoName(repo_url)
            branch = os.getenv(cls.BITBUCKET_BRANCH_ENV_VAR_NAME)
            commit_hash = os.getenv(cls.BITBUCKET_COMMIT_ENV_VAR_NAME)
            commit_message = GitHelper.get_commit_message()
            
            if not branch:
                branch = GitHelper.get_branch()

            if not commit_hash:
                commit_hash = GitHelper.get_commit_hash()

            test_run_id = cls.get_test_run_id(repo_url, commit_hash)

            env_info = EnvironmentInfo(test_run_id, cls.ENVIRONMENT, repo_url, repo_name, 
                branch, commit_hash, commit_message)
            print_debug_message_to_console("Bitbucket Environment info: {}".format(env_info.to_json()))
            return env_info
        except Exception as err:
            print_debug_message_to_console("Unable to build environment info: {}".format(err))
            LOGGER.error("Unable to build environment info: {}".format(err))
            pass
        return None


from __future__ import absolute_import
import wrapt

from thundra import utils, constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.django import DjangoORMIntegration

try:
    from django.conf import settings
except ImportError:
    settings = None

THUNDRA_MIDDLEWARE = "thundra.wrappers.django.middleware.ThundraMiddleware"

django_orm_integration = DjangoORMIntegration()


def _wrapper(wrapped, instance, args, kwargs):
    try:
        if getattr(settings, 'MIDDLEWARE', None):
            if THUNDRA_MIDDLEWARE in settings.MIDDLEWARE:
                return wrapped(*args, **kwargs)

            if isinstance(settings.MIDDLEWARE, tuple):
                settings.MIDDLEWARE = (THUNDRA_MIDDLEWARE,) + settings.MIDDLEWARE
            elif isinstance(settings.MIDDLEWARE, list):
                settings.MIDDLEWARE = [THUNDRA_MIDDLEWARE] + settings.MIDDLEWARE
        elif getattr(settings, 'MIDDLEWARE_CLASSES', None):
            if THUNDRA_MIDDLEWARE in settings.MIDDLEWARE_CLASSES:
                return wrapped(*args, **kwargs)

            if isinstance(settings.MIDDLEWARE_CLASSES, tuple):
                settings.MIDDLEWARE = (THUNDRA_MIDDLEWARE,) + settings.MIDDLEWARE_CLASSES
            elif isinstance(settings.MIDDLEWARE_CLASSES, list):
                settings.MIDDLEWARE = [THUNDRA_MIDDLEWARE] + settings.MIDDLEWARE_CLASSES

    except Exception:
        pass
    return wrapped(*args, **kwargs)


def db_execute_wrapper(execute, sql, params, many, context):
    return django_orm_integration.run_and_trace(
        execute,
        None,
        [sql,
         params,
         many,
         context],
        {}
    )


def install_db_execute_wrapper(connection, **kwargs):
    if db_execute_wrapper not in connection.execute_wrappers:
        connection.execute_wrappers.insert(0, db_execute_wrapper)


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_DJANGO_DISABLE) and (
            not utils.get_env_variable(constants.AWS_LAMBDA_FUNCTION_NAME)):
        wrapt.wrap_function_wrapper(
            'django.core.handlers.base',
            'BaseHandler.load_middleware',
            _wrapper
        )

    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_DJANGO_ORM_DISABLE):
        try:
            from django import VERSION
            from django.db import connections
            from django.db.backends.signals import connection_created

            if VERSION >= (2, 0):
                for connection in connections.all():
                    install_db_execute_wrapper(connection)

                connection_created.connect(install_db_execute_wrapper)
        except:
            pass



from .test_run_context import TestRunContext
from .test_run_finish import TestRunFinish
from .test_run_monitoring import TestRunMonitoring
from .test_run_result import TestRunResult
from .test_run_start import TestRunStart
from .test_run_status import TestRunStatus
from .terminator import Terminator, ThreadExecutorTerminator

__all__ = [
    "TestRunContext",
    "TestRunFinish",
    "TestRunMonitoring",
    "TestRunResult",
    "TestRunStart",
    "TestRunStatus",
    "Terminator",
    "ThreadExecutorTerminator"
]


from thundra import constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.base_integration import BaseIntegration


class RedisIntegration(BaseIntegration):
    CLASS_TYPE = 'redis'

    def __init__(self):
        pass

    def get_operation_name(self, wrapped, instance, args, kwargs):
        connection_kwargs = instance.connection_pool.connection_kwargs
        return connection_kwargs.get('host', 'Redis')

    def before_call(self, scope, wrapped, instance, args, kwargs, response, exception):
        connection_kwargs = instance.connection_pool.connection_kwargs
        host = connection_kwargs.get('host', '')
        port = connection_kwargs.get('port', '6379')
        command_type = wrapped.__name__.upper() or ""
        operation_type = constants.RedisCommandTypes.get(command_type, '')
        command = '{} {}'.format(command_type, ' '.join([str(arg) for arg in args]))

        scope.span.domain_name = constants.DomainNames['CACHE']
        scope.span.class_name = constants.ClassNames['REDIS']

        tags = {
            constants.SpanTags['OPERATION_TYPE']: operation_type,
            constants.DBTags['DB_INSTANCE']: host,
            constants.DBTags['DB_STATEMENT_TYPE']: operation_type,
            constants.DBTags['DB_TYPE']: 'redis',
            constants.RedisTags['REDIS_HOST']: host,
            constants.RedisTags['REDIS_PORT']: port,
            constants.RedisTags['REDIS_COMMAND_TYPE']: command_type,
            constants.SpanTags['TOPOLOGY_VERTEX']: True,
        }

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_REDIS_COMMAND_MASK):
            tags[constants.DBTags['DB_STATEMENT']] = command
            tags[constants.RedisTags['REDIS_COMMAND']] = command

        scope.span.tags = tags



from thundra.context.execution_context import ExecutionContext
from foresight.test_runner_tags import TestRunnerTags
from foresight.model import TestRunContext


class TestSuiteExecutionContext(TestRunContext, ExecutionContext):

    def __init__(self, **opts):
        super(TestSuiteExecutionContext, self).__init__(**opts)
        super(TestRunContext, self).__init__(**opts)
        self.test_suite_name = opts.get("node_id", '')
        self.completed = False


    def get_operation_name(self):
        return "TEST_SUITE"


    def get_additional_start_tags(self):
        from foresight.test_runner_support import TestRunnerSupport
        test_run_scope = TestRunnerSupport.test_run_scope
        return {
            TestRunnerTags.TEST_RUN_ID: test_run_scope.id,
            TestRunnerTags.TEST_RUN_TASK_ID: test_run_scope.task_id,
            TestRunnerTags.TEST_SUITE: self.test_suite_name
        }


    def get_additional_finish_tags(self):
        return {
            TestRunnerTags.TEST_SUITE_FAILED_COUNT: self.failed_count.value,
            TestRunnerTags.TEST_SUITE_TOTAL_COUNT: self.total_count.value,
            TestRunnerTags.TEST_SUITE_ABORTED_COUNT: self.aborted_count.value,
            TestRunnerTags.TEST_SUITE_SKIPPED_COUNT: self.ignored_count.value,
            TestRunnerTags.TEST_SUITE_SUCCESSFUL_COUNT: self.successful_count.value,
            TestRunnerTags.TEST_TIMEOUT: self.timeout
        }


import uuid

from thundra import constants

_common_fields_list = [
    'agentVersion',
    'dataModelVersion',
    'applicationId',
    'applicationDomainName',
    'applicationClassName',
    'applicationName',
    'applicationVersion',
    'applicationStage',
    'applicationRuntime',
    'applicationRuntimeVersion',
    'applicationTags'
]


def init_composite_data_common_fields(data):
    return {field: data.get(field) for field in _common_fields_list}


def remove_common_fields(data):
    without_common_fields = data.copy()
    for field in _common_fields_list:
        try:
            del without_common_fields[field]
        except (KeyError, TypeError):
            pass
    return without_common_fields


def get_composite_data(all_monitoring_data, api_key, data):
    composite_data = {
        "type": "Composite", "dataModelVersion": constants.DATA_FORMAT_VERSION,
        "apiKey": api_key,
        'data': data
    }
    composite_data['data']['id'] = str(uuid.uuid4())
    composite_data['data']['type'] = "Composite"
    composite_data['data']['allMonitoringData'] = all_monitoring_data
    return composite_data



import abc
import logging
import time
from concurrent import futures

from thundra.compat import PY2, queue
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.plugins.config.thundra_config import ThundraConfig
from thundra.reporter import Reporter
from foresight.model import  Terminator

ABC = abc.ABCMeta('ABC', (object,), {})

logger = logging.getLogger(__name__)


class ThreadPoolExecutorWithQueueSizeLimit(futures.ThreadPoolExecutor):
    def __init__(self, maxsize=10000, *args, **kwargs):
        super(ThreadPoolExecutorWithQueueSizeLimit, self).__init__(*args, **kwargs)
        self._work_queue = NoWaitQueue(maxsize=maxsize)


class NoWaitQueue(queue.Queue):
    def __init__(self, *args, **kwargs):
        queue.Queue.__init__(self, *args, **kwargs)

    def put(self, item, **kwargs):
        queue.Queue.put(self, item, block=False)


class BaseWrapper(ABC):

    def __init__(self, api_key=None, disable_trace=False, disable_metric=True, disable_log=True, opts=None):
        self.plugins = None
        self.config = ThundraConfig()
        if opts is not None:
            self.config = ThundraConfig(opts)

        self.api_key = ConfigProvider.get(config_names.THUNDRA_APIKEY, api_key)
        if not self.api_key:
            self.api_key = self.config.api_key
        if self.api_key is None:
            logger.error('Please set "thundra_apiKey" from environment variables in order to use Thundra')

        self.reporter = Reporter(self.api_key)
        self.debugger_process = None

        if not ConfigProvider.get(config_names.THUNDRA_TRACE_INSTRUMENT_DISABLE):
            if not PY2:
                from thundra.plugins.trace.patcher import ImportPatcher
                self.import_patcher = ImportPatcher()
        self.thread_pool_executor = ThreadPoolExecutorWithQueueSizeLimit()
        terminator = Terminator()
        terminator.register_task(self)


    def execute_hook(self, name, data):
        if name == 'after:invocation':
            [plugin.hooks[name](data) for plugin in reversed(self.plugins) if hasattr(plugin, 'hooks')
             and name in plugin.hooks]
        else:
            [plugin.hooks[name](data) for plugin in self.plugins if
             hasattr(plugin, 'hooks') and name in plugin.hooks]


    def prepare_and_send_reports(self, execution_context):
        execution_context.finish_timestamp = int(time.time() * 1000)
        self.execute_hook('after:invocation', execution_context)
        self.reporter.send_reports(execution_context.reports)


    def prepare_and_send_reports_async(self, execution_context):
        execution_context.finish_timestamp = int(time.time() * 1000)
        self.execute_hook('after:invocation', execution_context)
        try:
            self.thread_pool_executor.submit(self.reporter.send_reports, execution_context.reports)
        except queue.Full:
            logger.error("Dropping the monitoring report as the request queue is full")
        except Exception as e:
            logger.error(e)


from __future__ import unicode_literals

import base64
import hashlib

import simplejson as json

from thundra import constants, utils
from thundra.compat import str
from thundra.plugins.invocation import invocation_support, invocation_trace_support

try:
    from BytesIO import BytesIO
except ImportError:
    from io import BytesIO
from gzip import GzipFile


class LambdaEventType:
    Kinesis = 'kinesis',
    Firehose = 'firehose',
    DynamoDB = 'dynamodb',
    SNS = 'sns',
    SQS = 'sqs',
    S3 = 's3',
    CloudWatchSchedule = 'cloudWatchSchedule',
    CloudWatchLogs = 'cloudWatchLogs',
    CloudFront = 'cloudFront',
    APIGatewayProxy = 'apiGatewayProxy',
    APIGateway = 'apiGateway'
    Lambda = 'lambda'
    EventBridge = 'eventBridge'


def get_lambda_event_type(original_event, original_context):
    if 'Records' in original_event:
        if isinstance(original_event['Records'], list):
            records = original_event['Records'][0] or None
            if records is not None:
                if 'dynamodb' in records and records["eventSource"] == "aws:dynamodb":
                    return LambdaEventType.DynamoDB
                if 'kinesis' in records and records['eventSource'] == "aws:kinesis":
                    return LambdaEventType.Kinesis
                if 'EventSource' in records and records['EventSource'] == "aws:sns":
                    return LambdaEventType.SNS
                if 'eventSource' in records and records['eventSource'] == "aws:sqs":
                    return LambdaEventType.SQS
                if 's3' in records:
                    return LambdaEventType.S3
                if 'cf' in records:
                    return LambdaEventType.CloudFront

    elif 'detail-type' in original_event and original_event['detail-type'] == 'Scheduled Event' and \
            isinstance(original_event.get('resources'), list):
        return LambdaEventType.CloudWatchSchedule

    elif 'awslogs' in original_event and 'data' in original_event['awslogs']:
        return LambdaEventType.CloudWatchLogs

    elif 'deliveryStreamArn' in original_event and isinstance(original_event['records'], list):
        return LambdaEventType.Firehose

    elif 'requestContext' in original_event and 'headers' in original_event:
        return LambdaEventType.APIGatewayProxy

    elif 'context' in original_event and 'params' in original_event and 'header' in original_event['params']:
        return LambdaEventType.APIGateway

    elif 'detail-type' in original_event and 'detail' in original_event and \
            isinstance(original_event.get('resources'), list):
        return LambdaEventType.EventBridge

    elif 'client_context' in vars(original_context):
        return LambdaEventType.Lambda


def inject_trigger_tags_for_kinesis(span, original_event):
    domain_name = constants.DomainNames['STREAM']
    class_name = constants.ClassNames['KINESIS']
    stream_names = []
    trace_links = []
    for record in original_event['Records']:
        event_source_arn = record['eventSourceARN']
        index = event_source_arn.index('/') + 1
        stream_name = event_source_arn[index:]
        stream_names.append(stream_name)

        region = record['awsRegion']
        if "eventID" in record:
            trace_links.append(region + ':' + stream_name + ':' + record["eventID"])
    operation_names = list(set(stream_names))

    invocation_trace_support.add_incoming_trace_links(trace_links)

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_firehose(span, original_event):
    domain_name = constants.DomainNames['STREAM']
    class_name = constants.ClassNames['FIREHOSE']
    stream_arn = original_event['deliveryStreamArn']
    index = stream_arn.index('/') + 1
    stream_name = stream_arn[index:]
    operation_names = [stream_name]

    region = original_event['region']

    trace_links = []
    if 'records' in original_event:
        for record in original_event['records']:
            if "approximateArrivalTimestamp" in record and "data" in record:
                timestamp = record["approximateArrivalTimestamp"] / 1000
                try:
                    data = base64.b64decode(record["data"])
                    data_md5 = hashlib.md5(data).hexdigest()

                    trace_links.append(region + ':' + stream_name + ':' + str(int(timestamp - 1)) + ':' + data_md5)
                    trace_links.append(region + ':' + stream_name + ':' + str(int(timestamp)) + ':' + data_md5)
                    trace_links.append(region + ':' + stream_name + ':' + str(int(timestamp + 1)) + ':' + data_md5)

                except Exception:
                    pass

    invocation_trace_support.add_incoming_trace_links(trace_links)

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_dynamodb(span, original_event):
    domain_name = constants.DomainNames['DB']
    class_name = constants.ClassNames['DYNAMODB']
    table_names = []
    trace_links = []
    for record in original_event['Records']:
        table_name = record['eventSourceARN'].split('/')[1]
        table_names.append(table_name)

        region = record['awsRegion']

        trace_link_found = False
        if record['eventName'] == "INSERT" or record['eventName'] == "MODIFY":
            new_image = record['dynamodb'].get('NewImage')
            if new_image and new_image.get(constants.THUNDRA_SPAN_ID_KEY):
                span_id = new_image.get(constants.THUNDRA_SPAN_ID_KEY).get('S')
                trace_link_found = True
                trace_links.append("SAVE:" + span_id)

        elif record['eventName'] == "REMOVE":
            old_image = record['dynamodb'].get('OldImage')
            if old_image and old_image.get(constants.THUNDRA_SPAN_ID_KEY):
                span_id = old_image.get(constants.THUNDRA_SPAN_ID_KEY).get('S')
                trace_link_found = True
                trace_links.append("DELETE:" + span_id)

        if not trace_link_found:
            creation_time = record['dynamodb'].get('ApproximateCreationDateTime')
            if creation_time:
                if record['eventName'] == "INSERT" or record['eventName'] == "MODIFY":
                    add_dynamodb_trace_links(trace_links, region, table_name, creation_time, "SAVE",
                                             record['dynamodb'].get('NewImage'))
                    add_dynamodb_trace_links(trace_links, region, table_name, creation_time, "SAVE",
                                             record['dynamodb'].get('Keys'))

                elif record['eventName'] == "REMOVE":
                    add_dynamodb_trace_links(trace_links, region, table_name, creation_time, "DELETE",
                                             record['dynamodb'].get('Keys'))

    invocation_trace_support.add_incoming_trace_links(trace_links)

    operation_names = list(set(table_names))
    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def add_dynamodb_trace_links(trace_links, region, table_name, creation_time, operation_type, attributes):
    timestamp = creation_time - 1

    if attributes:
        attributes_hash = hashlib.md5(attributes_to_str(attributes).encode()).hexdigest()

    if attributes_hash:
        for i in range(3):
            trace_links.append(region + ':' + table_name + ':' + str(
                int(timestamp + i)) + ':' + operation_type + ':' + attributes_hash)


def attributes_to_str(attributes):
    sorted_keys = sorted(attributes.keys())
    attributes_sorted = []
    for attr in sorted_keys:
        try:
            key = list(attributes[attr].keys())[0]
            attributes_sorted.append(attr + '=' + '{' + key + ': ' + str(attributes[attr][key]) + '}')
        except Exception:
            pass
    return ', '.join(attributes_sorted)


def inject_trigger_tags_for_sns(span, original_event):
    domain_name = constants.DomainNames['MESSAGING']
    class_name = constants.ClassNames['SNS']
    topic_names = []
    trace_links = []
    for record in original_event['Records']:
        topic_arn = record['Sns']['TopicArn']
        topic_name = topic_arn.split(':')[-1]
        topic_names.append(topic_name)
        if 'MessageId' in record['Sns']:
            trace_links.append(record['Sns']['MessageId'])
    operation_names = list(set(topic_names))

    invocation_trace_support.add_incoming_trace_links(trace_links)

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_sqs(span, original_event):
    domain_name = constants.DomainNames['MESSAGING']
    class_name = constants.ClassNames['SQS']
    queue_names = []
    trace_links = []
    for record in original_event['Records']:
        queue_arn = record['eventSourceARN']
        queue_name = queue_arn.split(':')[-1]
        queue_names.append(queue_name)
        trace_links.append(record['messageId'])
    operation_names = list(set(queue_names))

    invocation_trace_support.add_incoming_trace_links(trace_links)

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_s3(span, original_event):
    domain_name = constants.DomainNames['STORAGE']
    class_name = constants.ClassNames['S3']
    bucket_names = []
    trace_links = []
    for record in original_event['Records']:
        if "responseElements" in record and "x-amz-request-id" in record["responseElements"]:
            trace_links.append(record["responseElements"]["x-amz-request-id"])
        bucket_names.append(record['s3']['bucket']['name'])
    operation_names = list(set(bucket_names))

    invocation_trace_support.add_incoming_trace_links(trace_links)

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_cloudwatch_schedule(span, original_event):
    domain_name = constants.DomainNames['SCHEDULE']
    class_name = constants.ClassNames['SCHEDULE']
    schedule_names = []
    for resource in original_event['resources']:
        schedule_names.append(resource.split('/')[-1])
    operation_names = list(set(schedule_names))

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_cloudwatch_logs(span, original_event):
    domain_name = constants.DomainNames['LOG']
    class_name = constants.ClassNames['CLOUDWATCHLOG']
    operation_names = []
    try:
        compressed_data = base64.b64decode(original_event['awslogs']['data'])
        decompressed_data = json.loads(str(GzipFile(fileobj=BytesIO(compressed_data)).read(), 'utf-8'))
        operation_names = [decompressed_data['logGroup']]
    except Exception as e:
        print('Error handling base64 format!', e)

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_cloudfront(span, original_event):
    domain_name = constants.DomainNames['CDN']
    class_name = constants.ClassNames['CLOUDFRONT']
    uris = []
    for record in original_event['Records']:
        try:
            uris.append(record['cf']['request']['uri'])
        except:
            pass
    operation_names = list(set(uris))

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_api_gateway_proxy(span, original_event):
    domain_name = constants.DomainNames['API']
    class_name = constants.ClassNames['APIGATEWAY']

    operation_names = []
    resource = utils.extract_api_gw_resource_name(original_event)
    if resource:
        operation_names.append(resource)
        invocation_support.set_application_resource_name(resource)

    if original_event.get('headers') and constants.THUNDRA_SPAN_ID_KEY in original_event['headers']:
        invocation_trace_support.add_incoming_trace_links([original_event['headers'][constants.THUNDRA_SPAN_ID_KEY]])

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_api_gateway(span, original_event):
    domain_name = constants.DomainNames['API']
    class_name = constants.ClassNames['APIGATEWAY']
    path = '/' + str(original_event['context']['stage']) + str(original_event['params']['path'])
    operation_names = [str(original_event['params']['header']['Host']) + path]

    invocation_support.set_application_resource_name(path)
    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def inject_trigger_tags_for_lambda(span, original_context):
    try:
        if original_context:
            if 'client_context' in vars(original_context):
                if original_context.client_context:
                    if original_context.client_context.custom:
                        domain_name = constants.DomainNames['API']
                        class_name = constants.ClassNames['LAMBDA']
                        operation_names = [original_context.client_context.custom[constants.TRIGGER_OPERATION_NAME_KEY]]

                        inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
                        inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)

        if 'aws_request_id' in vars(original_context):
            invocation_trace_support.add_incoming_trace_links([original_context.aws_request_id])
    except Exception as e:
        pass


def inject_trigger_tags_for_eventbridge(span, original_event):
    domain_name = constants.DomainNames['MESSAGING']
    class_name = constants.ClassNames['EVENTBRIDGE']
    operation_names = [original_event['detail-type']]

    invocation_trace_support.add_incoming_trace_links([original_event['id']])

    inject_trigger_tags_to_span(span, domain_name, class_name, operation_names)
    inject_trigger_tags_to_invocation(domain_name, class_name, operation_names)


def extract_trace_link_from_event(original_event):
    try:
        if '_thundra' in original_event:
            invocation_trace_support.add_incoming_trace_links([original_event['_thundra']['trace_link']])
    except Exception:
        pass


def inject_trigger_tags_to_span(span, domain_name, class_name, operation_names):
    span.set_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME'], domain_name)
    span.set_tag(constants.SpanTags['TRIGGER_CLASS_NAME'], class_name)
    span.set_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES'], operation_names)


def inject_trigger_tags_to_invocation(domain_name, class_name, operation_names):
    invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_DOMAIN_NAME'], domain_name)
    invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_CLASS_NAME'], class_name)
    invocation_support.set_agent_tag(constants.SpanTags['TRIGGER_OPERATION_NAMES'], operation_names)



import logging
import sys
from threading import Lock

import wrapt

from thundra import constants
from thundra.compat import PY37, PY38
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.plugins.config.log_config import LogConfig
from thundra.plugins.log import log_support
from thundra.plugins.log.thundra_log_handler import ThundraLogHandler
from thundra.plugins.log.thundra_logger import StreamToLogger


class LogPlugin:
    lock = Lock()
    wrapped = False

    def __init__(self, plugin_context=None, config=None):
        self.hooks = {
            'before:invocation': self.before_invocation,
            'after:invocation': self.after_invocation
        }
        self.plugin_context = plugin_context
        self.logger = logging.getLogger('STDOUT')

        if isinstance(config, LogConfig):
            self.config = config
        else:
            self.config = LogConfig()

        with LogPlugin.lock:
            if (not LogPlugin.wrapped) and (
                    not ConfigProvider.get(config_names.THUNDRA_LOG_CONSOLE_DISABLE)):
                if PY37 or PY38:
                    wrapt.wrap_function_wrapper(
                        'builtins',
                        'print',
                        self._wrapper
                    )
                else:
                    sys.stdout = StreamToLogger(self.logger, sys.stdout)
                LogPlugin.wrapped = True

        if not ConfigProvider.get(config_names.THUNDRA_LOG_CONSOLE_DISABLE):
            handler = ThundraLogHandler()
            has_thundra_log_handler = False
            for log_handlers in self.logger.handlers:
                if isinstance(log_handlers, ThundraLogHandler):
                    has_thundra_log_handler = True
            if not has_thundra_log_handler:
                self.logger.addHandler(handler)
            self.logger.setLevel(logging.INFO)
            handler.setLevel(logging.INFO)
            self.logger.propagate = False


    def _wrapper(self, wrapped, instance, args, kwargs):
        try:
            wrapped(*args, **kwargs)
            self.logger.info(str(args[0]))
        except:
            pass


    def before_invocation(self, execution_context):
        execution_context.capture_log = True


    def after_invocation(self, execution_context):
        execution_context.capture_log = False
        log_data = {
            'type': "Log",
            'agentVersion': constants.THUNDRA_AGENT_VERSION,
            'dataModelVersion': constants.DATA_FORMAT_VERSION,
            'traceId': execution_context.trace_id,
            'transactionId': execution_context.transaction_id,
        }

        for log in execution_context.logs:
            # Add application related data
            application_info = self.plugin_context.application_info
            log_data.update(application_info)
            log.update(log_data)
            if self.check_sampled(log):
                log_report = {
                    'data': log,
                    'type': 'Log',
                    'apiKey': self.plugin_context.api_key,
                    'dataModelVersion': constants.DATA_FORMAT_VERSION
                }
                execution_context.report(log_report)


    def check_sampled(self, log):
        sampler = log_support.get_sampler()
        if self.config.sampler:
            sampler = self.config.sampler
        sampled = True
        if sampler is not None:
            try:
                sampled = sampler.is_sampled(log)
            except Exception:
                pass
        return sampled



import time
from threading import Lock

from thundra import constants
from thundra.samplers.base_sampler import BaseSampler

from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names


class TimeAwareSampler(BaseSampler):

    def __init__(self, time_freq=None):
        freq_from_env = ConfigProvider.get(config_names.THUNDRA_SAMPLER_TIMEAWARE_TIMEFREQ, -1)
        if freq_from_env > 0:
            self.time_freq = freq_from_env
        elif time_freq is not None:
            self.time_freq = time_freq
        else:
            self.time_freq = constants.DEFAULT_METRIC_SAMPLING_TIME_FREQ
        self._latest_time = 0
        self._lock = Lock()

    def is_sampled(self, args=None):
        sampled = False
        with self._lock:
            current_time = 1000 * time.time()
            if current_time > self._latest_time + self.time_freq:
                self._latest_time = current_time
                sampled = True
        return sampled



import copy
from threading import Lock


class ThundraRecorder:

    def __init__(self):
        self._lock = Lock()
        self._spans = []

    def record(self, span):
        with self._lock:
            self._spans.append(span)

    def get_spans(self):
        return copy.copy(self._spans)

    def clear(self):
        self._spans = []



from thundra.plugins.config.log_config import LogConfig
from thundra.plugins.config.metric_config import MetricConfig
from thundra.plugins.config.trace_config import TraceConfig


class ThundraConfig:
    def __init__(self, opts=None):
        if opts is None:
            opts = {}

        self.api_key = opts.get('apiKey')
        self.disable_thundra = opts.get('disableThundra')
        self.trace_config = TraceConfig(opts.get('traceConfig'))
        self.metric_config = MetricConfig(opts.get('metricConfig'))
        self.log_config = LogConfig(opts.get('logConfig'))



from thundra.samplers import (
    TimeAwareSampler, CountAwareSampler,
    CompositeSampler
)

_sampler = CompositeSampler(
    samplers=[
        TimeAwareSampler(),
        CountAwareSampler()
    ],
    operator='or'
)


def get_sampler():
    return _sampler


def set_sampler(sampler):
    global _sampler
    _sampler = sampler



import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.mongodb import CommandTracer


def _wrapper(wrapped, instance, args, kwargs):
    event_listeners = list(kwargs.pop('event_listeners', []))
    event_listeners.insert(0, CommandTracer())
    kwargs['event_listeners'] = event_listeners
    wrapped(*args, **kwargs)


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_MONGO_DISABLE):
        try:
            import pymongo.monitoring
            from bson.json_util import dumps
            wrapt.wrap_function_wrapper(
                'pymongo',
                'MongoClient.__init__',
                _wrapper
            )
        except:
            pass



import logging
from types import SimpleNamespace

import aiohttp

from thundra import constants, utils
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.opentracing.tracer import ThundraTracer

logger = logging.getLogger(__name__)


async def on_request_start(session, trace_config_ctx, params):
    tracer = ThundraTracer.get_instance()
    if not tracer.get_active_span():
        return

    url_dict = utils.parse_http_url(str(params.url),
                                    ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_URL_DEPTH))
    scope = tracer.start_active_span(operation_name=url_dict.get('operation_name'), finish_on_close=False)
    scope.span.class_name = constants.ClassNames['HTTP']
    scope.span.domain_name = constants.DomainNames['API']

    tags = {
        constants.SpanTags['OPERATION_TYPE']: params.method,
        constants.HttpTags['HTTP_METHOD']: params.method,
        constants.HttpTags['HTTP_URL']: url_dict.get('url'),
        constants.HttpTags['HTTP_PATH']: url_dict.get('path'),
        constants.HttpTags['HTTP_HOST']: url_dict.get('host'),
        constants.HttpTags['QUERY_PARAMS']: url_dict.get('query'),
        constants.SpanTags['TOPOLOGY_VERTEX']: True,
    }

    scope.span.tags = tags
    try:
        scope.span.on_started()
        try:
            params.headers.update({constants.THUNDRA_SPAN_ID_KEY: scope.span.span_id})
            scope.span.set_tag(constants.SpanTags['TRACE_LINKS'], [scope.span.span_id])
        except Exception as e:
            pass
        trace_config_ctx.scope = scope
    except Exception as e:
        logger.error(e)


async def on_request_chunk_sent(session, trace_config_ctx, params):
    if not hasattr(trace_config_ctx, "scope"):
        return
    scope = trace_config_ctx.scope
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_BODY_MASK) and \
            (scope.span.get_tag(constants.HttpTags["BODY"]) is None):
        body = params.chunk if params.chunk else ""
        scope.span.set_tag(constants.HttpTags["BODY"], body)


async def on_request_end(session, trace_config_ctx, params):
    if not hasattr(trace_config_ctx, "scope"):
        return
    scope = trace_config_ctx.scope

    response = params.response
    if response is not None:
        status_code = response.status
        scope.span.set_tag(constants.HttpTags['HTTP_STATUS'], status_code)

        if response.headers and (response.headers.get("x-amz-apigw-id") or response.headers.get("apigw-requestid")):
            scope.span.class_name = constants.ClassNames['APIGATEWAY']

        if response.headers and response.headers.get(constants.TRIGGER_RESOURCE_NAME_KEY):
            resource_name = response.headers.get(constants.TRIGGER_RESOURCE_NAME_KEY)
            scope.span.operation_name = resource_name

        if (status_code and ConfigProvider.get(
                config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_ERROR_STATUS_CODE_MIN) <= status_code):
            scope.span.set_tag('error.kind', "HttpError")
            scope.span.set_tag('error', True)
            scope.span.set_tag('error.message', response.reason)
    try:
        scope.span.finish()
    except Exception as e:
        logger.error(e)

    scope.close()


async def on_request_exception(session, trace_config_ctx, params):
    if not hasattr(trace_config_ctx, "scope"):
        return
    scope = trace_config_ctx.scope
    scope.span.set_error_to_tag(params.exception)
    try:
        scope.span.finish()
    except Exception as e:
        logger.error(e)

    scope.close()


def ThundraTraceConfig():
    """
    :returns: TraceConfig.
    """

    def _trace_config_ctx_factory(trace_request_ctx):
        return SimpleNamespace(
            trace_request_ctx=trace_request_ctx
        )

    trace_config = aiohttp.TraceConfig(trace_config_ctx_factory=_trace_config_ctx_factory)
    trace_config.on_request_start.append(on_request_start)
    trace_config.on_request_end.append(on_request_end)
    trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
    trace_config.on_request_exception.append(on_request_exception)
    return trace_config



import logging

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider

loggers = {}


class StreamToLogger(object):
    def __init__(self, logger, stdout):
        self.logger = logger
        self.stdout = stdout

    def write(self, buf):
        for line in buf.rstrip().splitlines():
            self.logger.info(line.rstrip())
        self.stdout.write(buf)

    def flush(self):
        self.stdout.flush()


def get_logger(name):
    global loggers
    if loggers.get(name):
        return loggers.get(name)
    else:
        format = "%(asctime)s  - %(levelname)s - %(name)s - %(message)s"
        if name is None:
            logger = logging.getLogger(__name__)
        else:
            logger = logging.getLogger(name)
        logger.setLevel(logging.DEBUG)
        console_handler = logging.StreamHandler()
        console_handler.setLevel(logging.DEBUG)
        ch_format = logging.Formatter(format)
        console_handler.setFormatter(ch_format)
        logger.addHandler(console_handler)
        loggers[name] = logger
        return logger


def log_to_console(message, handler):
    logger = get_logger(handler)
    logging.getLogger().handlers = []
    logger.debug(message)


def debug_logger(msg, handler=None):
    if ConfigProvider.get(config_names.THUNDRA_DEBUG_ENABLE):
        if hasattr(msg, '__dict__'):
            log_to_console(msg, handler)
            display = vars(msg)
            log_to_console(display, handler)
            for key, _ in display.items():
                debug_logger_helper(getattr(msg, key), handler)
        else:
            log_to_console(msg, handler)


def debug_logger_helper(msg, handler):
    if hasattr(msg, '__dict__'):
        log_to_console(msg, handler)
        display = vars(msg)
        log_to_console(display, handler)
        for key, _ in display.items():
            debug_logger_helper(getattr(msg, key), handler)




class RdbBaseIntegration():
    _OPERATION_TO_TABLE_NAME_KEYWORD = {
        'select': 'from',
        'insert': 'into',
        'update': 'update',
        'delete': 'from',
        'create': 'table'
    }

    _OPERATION_TO_TYPE = {
        'select': 'READ',
        'insert': 'WRITE',
        'update': 'WRITE',
        'delete': 'DELETE'
    }



import wrapt

from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.integrations.base_integration import BaseIntegration
from thundra.integrations.modules.requests import _wrapper as request_wrapper

import thundra.integrations.botocore

INTEGRATIONS = {
    class_obj.CLASS_TYPE: class_obj()
    for class_obj in BaseIntegration.__subclasses__()
}


def _wrapper(wrapped, instance, args, kwargs):
    integration_name = instance.__class__.__name__.lower()

    if integration_name in INTEGRATIONS:
        return INTEGRATIONS[integration_name].run_and_trace(
            wrapped,
            instance,
            args,
            kwargs
        )

    return INTEGRATIONS['default'].run_and_trace(
        wrapped,
        instance,
        args,
        kwargs
    )


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_AWS_DISABLE):
        wrapt.wrap_function_wrapper(
            'botocore.client',
            'BaseClient._make_api_call',
            _wrapper
        )
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_HTTP_DISABLE):
        try:
            wrapt.wrap_function_wrapper(
                'botocore.vendored.requests',
                'Session.send',
                request_wrapper
            )
        except Exception:
            # Vendored version of requests is removed from botocore
            pass



import os, logging

logger = logging.getLogger(__name__)

# Platform independently user home folder path.
user_home_dir = os.path.expanduser("~")


def get_parent_dir(path):
    return os.path.abspath(os.path.join(path, os.pardir))


def backward_search_for_file(starting_path, filename_to_search):
    try:
        if starting_path and filename_to_search and starting_path != user_home_dir:
            current_folder_items = os.listdir(starting_path)
            if filename_to_search in current_folder_items:
                return starting_path
            starting_path = get_parent_dir(starting_path)
            return backward_search_for_file(starting_path, filename_to_search)
    except Exception as e:
        logger.error("backward_search_for_file error: {}".format(e))
        pass
    return None


import wrapt

from thundra import utils, constants
from thundra.config import config_names
from thundra.config.config_provider import ConfigProvider
from thundra.wrappers.tornado.middleware import ThundraMiddleware


def _init_wrapper(_wrapped, _application, args, kwargs):
    _wrapped(*args, **kwargs)

    middleware = _application.settings.get('_thundra_middleware')
    if middleware is None:
        _application.settings['_thundra_middleware'] = ThundraMiddleware()


async def _execute_wrapper(_wrapped, _handler, args, kwargs):
    middleware = _handler.settings.get('_thundra_middleware')
    middleware.execute(_handler)
    return await _wrapped(*args, **kwargs)


def _on_finish_wrapper(_wrapped, _handler, args, kwargs):
    middleware = _handler.settings.get('_thundra_middleware')
    middleware.finish(_handler)
    return _wrapped(*args, **kwargs)


def _log_exception_wrapper(_wrapped, _handler, args, kwargs):
    value = args[1] if len(args) == 3 else None
    if value is None:
        return _wrapped(*args, **kwargs)

    middleware = _handler.settings.get('_thundra_middleware')
    try:
        from tornado.web import HTTPError
        if not isinstance(value, HTTPError) or 500 <= value.status_code <= 599:
            middleware.finish(_handler, error=value)
    except ImportError:
        error = type('', (object,), {"status_code": 500})()
        middleware.finish(_handler, error=error)

    return _wrapped(*args, **kwargs)


def patch():
    if not ConfigProvider.get(config_names.THUNDRA_TRACE_INTEGRATIONS_TORNADO_DISABLE) and (
           not utils.get_env_variable(constants.AWS_LAMBDA_FUNCTION_NAME)):

        wrapt.wrap_function_wrapper(
            'tornado.web',
            'Application.__init__',
            _init_wrapper
        )
        wrapt.wrap_function_wrapper(
            'tornado.web',
            'RequestHandler._execute',
            _execute_wrapper
        )
        wrapt.wrap_function_wrapper(
            'tornado.web',
            'RequestHandler.on_finish',
            _on_finish_wrapper
        )
        wrapt.wrap_function_wrapper(
            'tornado.web',
            'RequestHandler.log_exception',
            _log_exception_wrapper
        )


import fastcounter

class TestRunContext:

    def __init__(self, total_count=0, successful_count = 0, failed_count=0, ignored_count=0, 
        aborted_count=0, **ignored):
        self.total_count = fastcounter.FastWriteCounter(total_count)
        self.successful_count = fastcounter.FastWriteCounter(successful_count)
        self.failed_count = fastcounter.FastWriteCounter(failed_count)
        self.ignored_count = fastcounter.FastWriteCounter(ignored_count)
        self.aborted_count = fastcounter.FastWriteCounter(aborted_count)

    def increase_total_count(self, count=1):
        self.total_count.increment(count)

    def increase_successful_count(self):
        self.successful_count.increment()

    def increase_failed_count(self):
        self.failed_count.increment()

    def increase_ignored_count(self, count=1):
        self.ignored_count.increment(count)

    def increase_aborted_count(self):
        self.aborted_count.increment()
    
    def get_total_count(self):
        return self.total_count.value

    def get_successful_count(self):
        return self.successful_count.value

    def get_failed_count(self):
        return self.failed_count.value

    def get_ignored_count(self):
        return self.ignored_count.value

    def get_aborted_count(self):
        return self.aborted_count.value


from foresight.constants.constants import ForesightConstants
from thundra.plugins.config.log_config import LogConfig
from thundra.wrappers.base_wrapper import BaseWrapper
import thundra.wrappers.wrapper_utils as wrapper_utils
from thundra.context.plugin_context import PluginContext
from thundra.application.global_application_info_provider import GlobalApplicationInfoProvider
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.context.tracing_execution_context_provider import TracingExecutionContextProvider
from foresight.context import (TestSuiteExecutionContext, TestCaseExecutionContext)
from foresight.sampler.max_count_aware_sampler import MaxCountAwareSampler 
from thundra.config.config_provider import ConfigProvider
from thundra.config import config_names
from foresight.constants.constants import ForesightConstants
import foresight.utils.generic_utils as utils
import logging

logger = logging.getLogger(__name__)

class TestWrapper(BaseWrapper):

    """
        Foresight wrapper util class. It keeps all the generic information for execution of test. 
        Creating test suite and test case, change application info, start and finish trace and invocation plugins.
    """

    __instance = None

    @staticmethod
    def get_instance(*args, **kwargs):
        return TestWrapper.__instance if TestWrapper.__instance else TestWrapper(*args, **kwargs) 


    def __init__(self, *, plugin_executor=None, api_key=None, disable_trace=False, disable_metric=True, disable_log=True, opts=None, **ignored):
        super(TestWrapper, self).__init__(api_key, disable_trace, disable_metric, disable_log, opts)
        ExecutionContextManager.set_provider(TracingExecutionContextProvider()) #TODO
        self.application_info_provider = GlobalApplicationInfoProvider()
        self._set_application_info("Foresight", "TestSuite", "TestSuite")
        self.plugin_context = PluginContext(application_info=self.application_info_provider.get_application_info(),
                                            request_count=0,
                                            executor=plugin_executor,
                                            api_key=self.api_key
        )
        max_test_log_count = ConfigProvider.get(config_names.THUNDRA_TEST_LOG_COUNT_MAX)
        self.config.log_config = LogConfig(sampler=MaxCountAwareSampler(max_test_log_count))
        self.plugins = wrapper_utils.initialize_plugins(self.plugin_context, disable_trace, disable_metric, disable_log,
                                                        config=self.config)
        TestWrapper.__instance = self


    def _set_application_info(self, application_class_name, application_domain_name, application_name):
        try:
            import pytest
            self.application_info_provider.update({
                "applicationClassName": application_class_name,
                "applicationDomainName": application_domain_name,
                "applicationName": application_name,
                "applicationVersion": pytest.__version__
            })
            self.application_info_provider.update({
                "applicationId": self._get_default_application_id()
            })
        except Exception as err:
            logger.error("Test wrapper utils application info set error: {}".format(err))
            pass


    def _get_default_application_id(self):
        try:
            application_info = self.application_info_provider.get_application_info()
            application_id = "python:test:pytest:{}:{}".format(application_info.get("applicationClassName"), application_info.get("applicationName"))
            return application_id.lower()
        except Exception as err:
            logger.error("get default application id error: {}".format(err))
            pass
        return "python:test:pytest:{}".format(utils.create_uuid4())


    def change_app_info(self, application_info):
        """It shall be used for when changing flow form test suite to test case or vice-versa.

        Args:
            application_info (ApplicationInfo): Thundra application info
        """
        try:
            self.application_info_provider.update(application_info.to_json())
        except Exception as err:
            logger.error("Test wrapper application info change error: {}".format(err))
            pass


    def create_test_suite_execution_context(self, test_suite_name=None):
        """ Creating test suite execution context

        Args:
            test_suite_name (str, optional): Unique test suite name. Defaults to None.

        Returns:
            TestSuiteExecutionContext: [description]
        """
        try:
            transaction_id = utils.create_uuid4()
            start_timestamp = utils.current_milli_time()
            return TestSuiteExecutionContext(
                transaction_id = transaction_id,
                start_timestamp = start_timestamp,
                node_id = test_suite_name
            )
        except Exception as err:
            logger.error("create test suite execution context error: {}".format(err))
            pass


    def create_test_case_execution_context(self, name, test_suite_name, test_case_id, app_info, parent_transaction_id=None):
        """Creating test case execution context

        Args:
            name (str): Test case name
            test_suite_name (str): unique test suite name
            test_case_id (str): unique test case id
            app_info (ApplicationInfo): Thundra application info
            parent_transaction_id (str, optional): Parent span transaction id for Thundra apm. Defaults to None.

        Returns:
            TestCaseExecutionContext: Store execution context for test case.
        """
        try:
            transaction_id = utils.create_uuid4()
            start_timestamp = utils.current_milli_time()
            test_class = app_info.application_class_name if app_info.application_class_name else None
            return TestCaseExecutionContext(
                transaction_id = transaction_id,
                start_timestamp = start_timestamp,
                test_suite_name = test_suite_name,
                node_id = test_case_id,
                parent_transaction_id = parent_transaction_id,
                name = name,
                method = ForesightConstants.TEST_OPERATION_NAME,
                test_class = test_class
            )
        except Exception as err:
            logger.error("create test case execution context error: {}".format(err))
            pass


    def start_trace(self, execution_context, tracer):
        """Calling after thundra trace pluging has been called by execute_hook. Create trace span.

        Args:
            execution_context (TestSuiteExecutionContext | TestSuiteExecutionContext): Execution context
            tracer (ThundraTracer): Thundra Tracer
        """
        try:
            operation_name = execution_context.get_operation_name()
            trace_id = utils.create_uuid4()
            scope = tracer.start_active_span(
                operation_name=operation_name,
                start_time=execution_context.start_timestamp,
                trace_id=trace_id,
                execution_context=execution_context,
                transaction_id=execution_context.transaction_id,
                finish_on_close=False,
            )
            root_span = scope.span
            app_info = self.application_info_provider.get_application_info()
            root_span.class_name = app_info.get("applicationClassName")
            root_span.domain_name = app_info.get("applicationDomainName")
            root_span.service_name = app_info.get("applicationName")
            execution_context.span_id = root_span.context.span_id
            execution_context.root_span = root_span
            execution_context.scope = scope
            execution_context.trace_id = trace_id
            
            root_span = execution_context.root_span
        except Exception as err:
            logger.error("Test wrapper start trace error: {}".format(err))
            pass


    def finish_trace(self, execution_context):
        """Finish trace.

        Args:
            execution_context (TestSuiteExecutionContext | TestSuiteExecutionContext): Execution context
        """
        try:
            root_span = execution_context.root_span
            scope = execution_context.scope
            try:
                root_span.finish(f_time=execution_context.finish_timestamp)
            except Exception:
                # TODO: handle root span finish errors
                pass
            finally:
                scope.close()
        except Exception as err:
            logger.error("test wrapper finish trace error: {}".format(err))
            pass


    def start_invocation(self, execution_context):
        """Calling after thundra invocation pluging has been called by execute_hook. Create trace span.

        Args:
            execution_context (TestSuiteExecutionContext | TestSuiteExecutionContext): Execution context
        """
        try:
            execution_context.invocation_data = wrapper_utils.create_invocation_data(self.plugin_context, execution_context)
        except Exception as err:
            logger.error("test wrapper start invocation error: {}".format(err))
            pass


    def finish_invocation(self, execution_context):
        """Finish invocation.

        Args:
            execution_context (TestSuiteExecutionContext | TestSuiteExecutionContext): Execution context
        """
        try:
            wrapper_utils.finish_invocation(execution_context)
        except Exception as err:
            logger.error("finish invocation error: {}".format(err))
            pass


    def before_test_process(self, execution_context):
        """Invoke trace and invocation plugings before each test suite and test case.

        Args:
            execution_context (TestSuiteExecutionContext | TestCaseExecutionContext): Thundra test execution context
        """
        try:
            self.execute_hook("before:invocation", execution_context)
        except Exception as err:
            logger.error("test wrapper before test process: {}".format(err))
            pass


    def after_test_process(self, execution_context):
        """Send collected data about test suit or test case after each finish.

        Args:
            execution_context (TestSuiteExecutionContext | TestCaseExecutionContext): Thundra test execution context
        """
        try:
            self.prepare_and_send_reports_async(execution_context)
        except Exception as err:
            logger.error("test wrapper after test process error: {}".format(err))
            pass


    def send_test_run_data(self, test_run_event):
        """Send test run data. test_run_monitoring data has been updated by application info because
         reporter has been used tih info whilst preparing composite data. 

        Args:
            test_run_event (TestRunStart, TestRunStatus, TestRunFinish): Tests event data
        """
        try:
            test_run_monitoring_data = test_run_event.get_monitoring_data()
            self.reporter.send_reports([test_run_monitoring_data], test_run_event = True)
        except Exception as err:
            logger.error("test wrapper send test run data error: {}".format(err))
            pass


from functools import wraps
import logging
from thundra import constants

from thundra.opentracing.tracer import ThundraTracer
from thundra.wrappers.base_wrapper import BaseWrapper

from thundra.config.config_provider import ConfigProvider
from thundra.context.execution_context_manager import ExecutionContextManager
from thundra.context.plugin_context import PluginContext
from thundra.application.global_application_info_provider import GlobalApplicationInfoProvider
from thundra.context.tracing_execution_context_provider import TracingExecutionContextProvider
from thundra.wrappers.fastapi import fastapi_executor
from thundra.config import config_names
from thundra.wrappers import wrapper_utils, web_wrapper_utils

import traceback


logger = logging.getLogger(__name__)


class FastapiWrapper(BaseWrapper):    
    __instance = None
    
    @staticmethod
    def get_instance():
        return FastapiWrapper() if FastapiWrapper.__instance is None else FastapiWrapper.__instance

    def __init__(self, api_key=None, disable_trace=False, disable_metric=True, disable_log=True, opts=None):
        super(FastapiWrapper, self).__init__(api_key, disable_trace, disable_metric, disable_log, opts)
        self.application_info_provider = GlobalApplicationInfoProvider()
        ExecutionContextManager.set_provider(TracingExecutionContextProvider())
        self.plugin_context = PluginContext(application_info=self.application_info_provider.get_application_info(),
                                            request_count=0,
                                            executor=fastapi_executor,
                                            api_key=self.api_key)

        self.plugins = wrapper_utils.initialize_plugins(self.plugin_context, disable_trace, disable_metric, disable_log,
                                                        config=self.config)

        web_wrapper_utils.update_application_info(self.application_info_provider, self.plugin_context.application_info,
                                                  constants.ClassNames['FASTAPI'])

        FastapiWrapper.__instance = self

    def before_request(self, request, req_body):
        execution_context = wrapper_utils.create_execution_context()
        execution_context.platform_data["request"] = request
        execution_context.platform_data["request"]["body"] = req_body

        self.plugin_context.request_count += 1
        self.execute_hook("before:invocation", execution_context)

        return execution_context


    def after_request(self, execution_context):
        self.prepare_and_send_reports_async(execution_context)

    def error_handler(self, error, execution_context):
        if error:
            execution_context.error = error

        self.prepare_and_send_reports_async(execution_context)
        
        

    def __call__(self, original_func):

        import inspect
        if hasattr(original_func, "_thundra_wrapped") or ConfigProvider.get(config_names.THUNDRA_DISABLE, False):
            return original_func

        @wraps(original_func)
        async def wrapper(*args, **kwargs):
            request = kwargs.get("request")
            if request is None or request.scope.get('_thundra_wrapped', False):
                if inspect.iscoroutinefunction(original_func):
                    return await original_func(*args, **kwargs)
                else:
                    return original_func(*args, **kwargs)

            request.scope['_thundra_wrapped'] = True
            try:
                req_body = request._body if hasattr(request, "_body") else None
                request.scope["thundra_execution_context"] = self.before_request(request.scope, req_body)
                execution_context = request.scope["thundra_execution_context"]
            except Exception as e:
                logger.error('Error during the before part of Thundra: {}'.format(e))
                if inspect.iscoroutinefunction(original_func):
                    return await original_func(*args, **kwargs)
                else:
                    return original_func(*args, **kwargs)

            response = None
            try:
                if inspect.iscoroutinefunction(original_func):
                    response = await original_func(*args, **kwargs)
                else:
                    response = original_func(*args, **kwargs)
            except Exception as e:
                try:
                    error = {
                        'type': type(e).__name__,
                        'message': str(e),
                        'traceback': traceback.format_exc()
                    }
                    self.error_handler(error, execution_context)
                except Exception as e_in:
                    logger.error("Error during the after part of Thundra: {}".format(e_in))
                raise e

            try:
                execution_context.response = response
                self.after_request(execution_context)
            except Exception as e:
                logger.error("Error during the after part of Thundra: {}".format(e))
            return response

        setattr(wrapper, '_thundra_wrapped', True)
        return wrapper